#!/bin/sh
# $Id: netams-linux-startup.sh,v 1.1.2.5 2006/06/02 19:33:48 anton Exp $
# netams        Startup shell script for NeTAMS 
# 
# V.N. Chernenko <vnc@vnc.org.ua>
#
### BEGIN INIT INFO
# Provides: netams
# Required-Start: $local_fs mysql apache2 
# Required-Stop:  $local_fs mysql 
# Default-Start:  3 5
# Default-Stop:   0 1 2 6
# Description:    NeTAMS (Network Traffic Accounting and Management Software) 
# 				Daemon for accounting network traffic data. 
# Short-Description: Start the netams daemon
### END INIT INFO

DAEMONDIR=/usr/sbin
CONFIG=/etc/netams/netams.cfg
PIDFILE=/var/run/netams.pid
NETAMS_BIN=$DAEMONDIR/netams
NETAMS_ARGS="-lf $CONFIG"

. /etc/rc.status

# First reset status of this service
rc_reset

# Return values acc. to LSB for all commands but status:
# 0 - success
# 1 - generic or unspecified error
# 2 - invalid or excess argument(s)
# 3 - unimplemented feature (e.g. "reload")
# 4 - insufficient privilege
# 5 - program is not installed
# 6 - program is not configured
# 7 - program is not running
# 
# Note that starting an already running service, stopping
# or restarting a not-running service as well as the restart
# with force-reload (in case signalling is not supported) are
# considered a success.

case "$1" in
    start)
        echo -n "Starting netams "
        ## Start daemon with startproc(8). If this fails
        ## the echo return value is set appropriate.

        # NOTE: startproc returns 0, even if service is 
        # already running to match LSB spec.
        startproc -q -p $PIDFILE $NETAMS_BIN $NETAMS_ARGS 

        # Remember status and be verbose
        rc_status -v
        ;;
    stop)
        echo -n "Shutting down netams "

        netamsctl shutdown > /dev/null
        rm -rf $PIDFILE

        # Remember status and be verbose
        rc_status -v
        ;;
    try-restart)
        ## Stop the service and if this succeeds (i.e. the 
        ## service was running before), start it again.
        ## Note: try-restart is not (yet) part of LSB (as of 0.7.5)
        $0 status >/dev/null &&  $0 restart

        # Remember status and be quiet
        rc_status
        ;;
    restart)
        ## Stop the service and regardless of whether it was
        ## running or not, start it again.
        $0 stop
        $0 start

        # Remember status and be quiet
        rc_status
        ;;
    status)
        echo -n "Checking for service spamd "
        ## Check status with checkproc(8), if process is running
        ## checkproc will return with exit status 0.

        # Return value is slightly different for the status command:
        # 0 - service running
        # 1 - service dead, but /var/run/  pid  file exists
        # 2 - service dead, but /var/lock/ lock file exists
        # 3 - service not running

        # NOTE: checkproc returns LSB compliant status values.
        checkproc -p $PIDFILE $NETAMS_BIN
        rc_status -v
        ;;
    *)
        echo "Usage: $0 {start|stop|status|try-restart|restart}"
        exit 1
        ;;
esac
rc_exit
