/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: ds_ulog.c,v 1.19.2.2 2006/03/02 11:44:34 anton Exp $ */

#if defined(LINUX) && !defined(IPTBL_NONE)

#include "netams.h"
#include "libipulog.h"

/////////////////////////////////////////////////////////////////////////////////////
void ds_ulog_cancel(void *ptr);
/////////////////////////////////////////////////////////////////////////////////////
void* ds_ulog(void *ss) {
	Service *s=(Service*)ss;
	ServiceDS_cfg *cfg=(ServiceDS_cfg*)s->cfg;
	
	struct ipulog_handle *ulog_h; /* our libipulog handle */
	ulog_packet_msg_t *m;
	FlowEngine *FE=cfg->FE;
        struct timeval start;
	unsigned char *packet=cfg->packet;
	int status;
	
	/* create ipulog handle */
        ulog_h = ipulog_create_handle(cfg->nlmask);
        if (!ulog_h) {
		aLog(D_ERR, "linux ulog handle failed: %s!\n", ipulog_strerror(ipulog_errno));
		return NULL;
	}
	SET_POLL(ulog_h->fd);

        pthread_cleanup_push(ds_ulog_cancel, (void*)ulog_h);
        pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

	aLog(D_INFO,"ULOG packet processing for data-source:%u initialized\n",s->instance);
        
	struct ipv4_key         key;
	entry                   *flow_entry;

	while(1) {
		CHECK_POLL(status);
			
		netams_gettimeofday(&start, NULL);
		FE->Expiresearch(&start);
		
		if(status==0) continue;

		status=ipulog_read(ulog_h, packet, MAX_PKT_SIZE, 1000);
		if(status <0) {
			aLog(D_ERR, "ds:%u : %s\n", cfg->id, ipulog_strerror(ipulog_errno));
			continue;
		}
		
		while((m=ipulog_get_packet(ulog_h, packet, status))) {
			IPv4GetKey((struct ip*) m->payload, &key);
			if(FE->Process((u_char *)&key, 1, ntohs(((struct ip*) m->payload)->ip_len), &flow_entry) == -1) {
				IPv4FillFlow(&key, flow_entry);
#ifdef LAYER7_FILTER
			if (cfg->layer7_detect!=LAYER7_DETECT_NONE) layer7_addinfo(key.tcp_info.dst_port, flow_entry);
#endif
			}

#ifdef LAYER7_FILTER
			if (cfg->layer7_detect!=LAYER7_DETECT_NONE) layer7_checkinfo(key.tcp_info.dst_port, flow_entry, (struct ip*) m->payload);
#endif
			sDSMeasure(cfg, &start, m->data_len);
		}
	}
	pthread_cleanup_pop(1);
	return NULL;
}
/////////////////////////////////////////////////////////////////////////////////////
void ds_ulog_cancel(void *ptr) {
	ipulog_destroy_handle((struct ipulog_handle*)ptr);
}
#endif
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
