/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: flow.h,v 1.44.2.2 2006/03/21 12:01:21 jura Exp $ */

//////////////////////////////////////////////////////////////////////////
#define MAX_IDX_SIZE		16

#define CHECK_EXPIRE 		ProcessorCfg.delay	//10	//seconds between expire checks
#define ACTIVE_TIMEOUT 		ProcessorCfg.lifetime	//60
#define INACTIVE_TIMEOUT	60 //

// bitmasks
#define ENTRY_EXPIRED		1
#define ENTRY_BLOCKED		2

#define FLOW_HASH_SIZE		0x10000
#define FLOW_HASH_MASK		(FLOW_HASH_SIZE-1)

//////////////////////////////////////////////////////////////////////////
typedef struct entry {
	u_char index[MAX_IDX_SIZE];

	class Flow *flow;
	
	struct  flow_info_value	*flow_info;
#ifdef LAYER7_FILTER
	struct layer7_info_value *layer7_info;  //to optimize layer7 operations
#endif
	u_char flags;   //EXPIRED,BLOCKED etc
#ifdef HAVE_BW	
	bwUlist *bwRoot;
#endif	
	u_short hash;
	entry *next;
	entry *next_active;
} entry;
//////////////////////////////////////////////////////////////////////////
typedef struct EngineMsg {
	EngineMsg 		*next;
	u_char 			freeSlot;
	struct attribute        *list;
} EngineMsg;
//////////////////////////////////////////////////////////////////////////
class FlowEngine {
	public:
		IPTree *IPtree;
		u_char instance;

#define MAX_CHUNK               2048 //protection against DoS - max chunks per hash node

#define FLOW_FW_CHECK   0x01
		u_char flags;
		
		entry **table;
		entry *ready;
		entry *active;

		struct timeval *tv; //we need this for bw 

		time_t t;
		time_t t_now;
		
		unsigned  e_used;
		unsigned  e_total;

		unsigned last_id, expired_id;
		unsigned long long sent_flows;
		
		DS_FIFO *fifo;
		EngineMsg *message;
		u_char max_flow_slots;

		FlowEngine(Service *s, u_char skip_fw_check);
		~FlowEngine();
		void CheckExpire(entry *e);
		void Expiresearch(struct timeval *tv);
		int Process(u_char *flow_key, u_long packets, u_long octets, entry **e2get);
		void FlushAll();
		void DoSend(entry *e);
		void DoSend(Flow *flow);
		
		u_char FW(entry *e);	
		u_char DSfw(NetUnit *u, entry *e, match mf);
#ifdef HAVE_BW
		BWEngine *BW;
#endif
};
//////////////////////////////////////////////////////////////////////////
struct ipv4_key {
	struct ipv4_info_value	ipv4_info;
	struct tcp_info_value	tcp_info;
	u_char			padding[2];
};

struct pcap_ipv4_key {
	struct ipv4_info_value	ipv4_info;
	struct tcp_info_value	tcp_info;
	u_short			dot1x;
};
//////////////////////////////////////////////////////////////////////////
struct ipv4_key* IPv4GetKey(struct ip *ip, struct ipv4_key *key);
void IPv4FillFlow(struct ipv4_key* ipv4_flow_key, entry *flow_entry);
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

