/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: main.c,v 1.76.2.1 2005/11/19 21:31:03 jura Exp $ */

#include "netams.h"

/////////////////////////////////////////////////////////////
void aMainInit();
void aMainRelease();
/////////////////////////////////////////////////////////////
Service *sMain;
int global_return_code=0;
int is_running=0;
language lang=LANG_EN;

int flag_nodaemon=0, flag_log=0, flag_quiet=0, sql_config=0, flag_pid=0, flag_nopid=0, flag_syslog=0;
FILE *LOGFILE;
FILE *configfile;
char *config_file_name=NULL;
char *pid_file_name=NULL;
struct timeval program_start;
time_t start_t;

/////////////////////////////////////////////////////////////
extern void CalibrateClock();
/////////////////////////////////////////////////////////////
int main(int argc, char **argv){
	int op;
	char buf[32];
	
	gettimeofday(&program_start, NULL);
	start_t 	= program_start.tv_sec;
	
	aMainInit();

	while((op=getopt(argc, argv, "qdlLPsf:p:")) != EOF){
		switch(op){
			case 'q': 
				flag_quiet = 1;
				break;
			case 'd': flag_nodaemon = 1;
				break;
			case 'l':
				flag_log = 1;
				flag_pid = 1;
				break;
			case 'L':
				flag_syslog = 1;
				flag_pid = 1;
				break;
			case 'P':
				flag_nopid = 1;
				break;
			case 'p':
				pid_file_name=set_string(optarg);
				flag_pid = 1;
				break;
			case 's':
				sql_config = 1;
				break;
			case 'f': 
				config_file_name=set_string(optarg);
				break;
			default: 
				break;
		}
	}

	if (flag_log) {
		LOGFILE = fopen(path_to_log, "at");
		if (LOGFILE==NULL) { 
			fprintf(stderr, "LOGFILE %s opening: %s\n", path_to_log, strerror(errno)); 
			aMainRelease();
			exit(-1); 
		}
		setvbuf(LOGFILE, NULL, _IOLBF, 0);
	}
	
	if (flag_syslog) openlog("netams", 0, LOG_DAEMON);
	if (!flag_quiet) aShowVersion(stdout);
	if (flag_log) aShowVersion(LOGFILE);

	if (config_file_name==NULL) 
		config_file_name=set_string((char*)path_to_config);
	
	if (!flag_nodaemon) {
		if(chdir(RUN_PATH)<0) {
			mkdir(RUN_PATH,S_IRWXU);
		}
		if(chdir(RUN_PATH)<0)
			aLog(D_CRIT, "Cannot chdir to directory %s: %s\n",RUN_PATH,strerror(errno));
		
		aLog(D_INFO, "Becoming a daemon...\n");
		daemon(1, !flag_quiet);
	} else
		aLog(D_INFO, "Stay in debug mode...\n");
	
	if (flag_pid && !flag_nopid) {
		if (pid_file_name==NULL) pid_file_name=set_string("/var/run/netams.pid");
		FILE *pidfile = fopen(pid_file_name, "wt");

		if (pidfile!=NULL) {
			fprintf(pidfile, "%u\n", getpid());
			fclose(pidfile);
		} else {
			fprintf(stderr, "Can't create pid-file: %s\n", strerror(errno));
		}
	}
        
	CalibrateClock();

	// handling signals
	signal(SIGHUP, (sig_t)logrotate);
	signal(SIGUSR1, (sig_t)logrotate);
	signal(SIGQUIT, (sig_t)termination);
	signal(SIGINT, (sig_t)termination);
	signal(SIGTERM, (sig_t)termination);

	aPermissionsInit();

	uInternal=new User("<internal>", 1);  uInternal->permissions=UPERM_ALL;
	Users->Insert(uInternal);

	
	cInternal->permissions=UPERM_ALL;
	cInternal->user=uInternal;
	Connections->Insert(cInternal);
	aDebugAdd(cInternal->debug, "none");

	sMain=new Service(SERVICE_MAIN, 0);
	Services->Insert(sMain);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	sMain->t_id=cInternal->t_id=pthread_self();

	if (sql_config) {
		if (cSqlBoot(cInternal, config_file_name)<0) {
			aMainRelease();
			exit(-1);
		}
	}
	else {
		if (cLoad(cInternal, config_file_name)<0) {
			aLog(D_CRIT, "main/configfile opening: %s\n", strerror(errno));
			aMainRelease();
			exit(-1);
		}
	}

	sMain->setSleep();

	Services->StartAll(sMain);

	sMain->Sleep(1); //we must sleep to be sure all services started

	if (!flag_quiet) 
		printf("%s is now operational, %s\n", aaa_fw_software_name, timeU2T(time(NULL), buf));
	
	sprintf(buf, "%u.%u.%u (build %u", aaa_fw_major_version, aaa_fw_minor_version, aaa_fw_subversion, aaa_fw_build_version);
	if (aaa_fw_build_version_local) sprintf(buf+strlen(buf), ".%d", aaa_fw_build_version_local);
	sprintf(buf+strlen(buf), ")");
	LogEvent(SYSTEM, 0, 0, 0, "NeTAMS Startup %s %s", buf, config_file_name);
	
	is_running=1;

	sMain->Sleep();

	// the last actual action
	is_running=0;
	
	/////////////////////////////////////////////////////////////
	// exit point
	if (!flag_quiet) 
		printf("Shutdown signal received at %s\n", timeU2T(time(NULL), buf));
	aLog(D_INFO, "Shutdown signal received at %s\n", timeU2T(time(NULL), buf));
	LogEvent(SYSTEM, 0, 0, 0, "NeTAMS Shutdown");

	Services->ShutdownAll(sMain);

	aFree(config_file_name); //we need this for "save"
	
	if (!flag_quiet) 
		printf("%s exiting at %s\n", aaa_fw_software_name, timeU2T(time(NULL), buf));
	aLog(D_INFO, "%s exiting with code %d\n", aaa_fw_software_name, global_return_code);
	aLog(D_INFO, "units locking summary:\ntries: %lu, failed: %lu, ratio: %3.2f\n", Units->tries_lock, Units->tries_lock_failed,  !Units->tries_lock?0:100*(float)((double)Units->tries_lock_failed/(double)Units->tries_lock));
	
	if (flag_pid && !flag_nopid && (pid_file_name!=NULL)) {
		unlink(pid_file_name);
		aFree(pid_file_name);
	}

	aMainRelease();
	
	if (flag_log) fclose(LOGFILE);

	if (flag_syslog) closelog();

	if( global_return_code == PARSE_RELOAD ) 
		if (-1==execvp(argv[0],argv)) perror("reload failed");
	
	return global_return_code;					  
}
/////////////////////////////////////////////////////////////
void aMainInit() {
	aDebugInit();
	cInternal=new Connection("<internal>", 1); 

	aMutexDebugInit();
	aMemoryDebugInit();

	Services        = new ServicesList();
	Connections     = new ConnectionsList();
	
	aOidsInit();
	
	MsgMgr 		= new MessageManager();
	PolicyL		= new PolicyList();
        Units  		= new NetUnitsList();
	Users  		= new UsersList();
}

void aMainRelease() {
	uInternal=NULL;
	delete Users;
	delete Units;
	delete PolicyL;
	delete MsgMgr;
	
	delete Services;	Services=NULL;
	delete sMain;
	delete Connections;     Connections=NULL;
	
	aMemoryDebugRelease();
	aMutexDebugRelease();

	delete cInternal;       cInternal=NULL;
}
/////////////////////////////////////////////////////////////
