/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: s_acl_server.c,v 1.13.2.4 2006/03/30 11:01:53 jura Exp $ */

#include "netams.h"

//////////////////////////////////////////////////////////////////////////
void *sAclServer(void *s);
void sAclServerCancel(void *);
unsigned long long sAclServer_ParseUptime(char *buffer);
int sAclServer_SendCommand(char *tx, char *rx, int max_rx, FILE **fd, ServiceAclServer_cfg *cfg, struct addrinfo *res);

//////////////////////////////////////////////////////////////////////////
void sAclServerInit(Service *s){
	s->cfg = (ServiceAclServer_cfg*)aMalloc(sizeof(ServiceAclServer_cfg));
	ServiceAclServer_cfg *cfg=(ServiceAclServer_cfg*)s->cfg;

	cfg->hostname=set_string("localhost");
	cfg->port=S_ACLSERVER_DEF_cmdport;
	cfg->direction=0;
	cfg->dynamic_name=NULL;
	cfg->acl_number=S_ACLSERVER_DEF_aclnumber;
	cfg->is_cisco=0;
	cfg->delay=S_ACLSERVER_DEF_delay;
	cfg->uptime=0;
	cfg->in=new FIFO(MAX_UNITS);
	
	pthread_create(&(s->t_id), NULL, &sAclServer, s);
}

//////////////////////////////////////////////////////////////////////////
void sAclServerProcessCfg(char *param[], Connection *conn, u_char no_flag){
	ServiceAclServer_cfg *cfg=(ServiceAclServer_cfg*)conn->service->cfg;

	if (!strcasecmp(param[0], "hostname") && param[1]!=empty) {
		if (cfg->hostname) aFree(cfg->hostname);
		cfg->hostname=set_string(param[1]);
		aParse(conn, "hostname is set to %s\n", cfg->hostname);
		if (param[2]!=empty) {
			cfg->port=strtol(param[2], NULL, 10);
			aParse(conn, "port is set to %u\n", cfg->port);
		}
	}
	else if (!strcasecmp(param[0], "direction")) {
		if (!strcasecmp(param[1], "src")) cfg->direction=0;
		else if (!strcasecmp(param[1], "dst")) cfg->direction=1;
		else aParse(conn, "cannot set acl direction: value invalid\n");
		aParse(conn, "acl direction is set to \'%s\'\n", cfg->direction?"dst":"src");
	}
	else if (!strcasecmp(param[0], "dynamic-name")) {
		if (param[1]!=empty) {
			cfg->dynamic_name=set_string(param[1]);
			aParse(conn, "dynamic acl name is set to %s\n", cfg->dynamic_name);
		}
		else {
			cfg->dynamic_name=NULL;
			aParse(conn, "dynamic acl name cleared\n");
		}
	}
	else if (!strcasecmp(param[0], "acl-number")) {
		if (param[1]!=empty) {
			cfg->acl_number=strtol(param[1], NULL, 10);
			if (!strcasecmp(param[2], "cisco")) cfg->is_cisco=1;
			else cfg->is_cisco=0;
			aParse(conn, "acl number is set to %u, device type is %scisco\n", cfg->acl_number, cfg->is_cisco?"":"non-");
		}
		else aParse(conn, "cannot set acl number\n");
	}
	 else if (!strcasecmp(param[0], "delay")) {
		cfg->delay=strtol(param[1], NULL, 10);
		aParse(conn, "acl check delay is set to %u\n", cfg->delay);
	}
	 else if (!strcasecmp(param[0], "set-uptime")) {
		cfg->uptime=strtoll(param[1], NULL, 10);
		aParse(conn, "acl uptime forcibly set to %llu\n", cfg->uptime);
	}
	else aParse(conn, "unknown AclServer command: %s\n", param[0]);
}
//////////////////////////////////////////////////////////////////////////
void *sAclServer(void *ss){
	Service *s=(Service*)ss;

	aLog(D_INFO, "service AclServer:%u thread started (%p)\n", s->instance, s->cfg);
	pthread_cleanup_push(sAclServerCancel, s);
	SET_CANCEL();
	s->t_id=pthread_self();

	ServiceAclServer_cfg *cfg=(ServiceAclServer_cfg*)s->cfg;
	
	// now we should sleep before starting AclServer service exec
	if(!(s->flags&SERVICE_RUN)) s->Sleep();

	// then wait a little to allow old data to be read by processor
	s->Sleep(ProcessorCfg.delay);
	
	struct addrinfo *res;
	struct addrinfo hints; 
	bzero(&hints, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	FILE *fd;
	char buffer_tx[256];
	char buffer_rx[4096];
	char ip_s[32];
	Message *msg;
	Message_AclServer *msgacl;
	unsigned msgs_proc, msgs_proc_ok;
	
	while(1) {
		if (cfg->uptime) s->Sleep(cfg->delay);

		cfg->uptime++;

		aDebug(DEBUG_ACLSERVER, "acl server checking every %u seconds\n", cfg->delay);
		
		sprintf(ip_s, "%u", cfg->port);
		int error=getaddrinfo(cfg->hostname, ip_s, &hints, &res);
		if (error) {
			aDebug(DEBUG_ACLSERVER, "acl getaddrinfo error: %s \n", gai_strerror(error));
			continue;
		}
		
		fd=NULL;
		if (sAclServer_SendCommand("show version | i uptime", buffer_rx, 1024*4, &fd, cfg, res)) {
			continue;
		}

		unsigned long long c_uptime = sAclServer_ParseUptime(buffer_rx);
		aDebug(DEBUG_ACLSERVER, "known: %llu, remote uptime: %s %llu \n", cfg->uptime, buffer_rx, c_uptime);
		if (cfg->uptime  > c_uptime + cfg->delay || cfg->uptime==1) {
			// 1.1. flush old list of rules
			// 1.2. install the new list for ALL units with !sys-allow

			netams_rwlock_rdlock(&Units->rwlock);
			for (NetUnit *u=Units->root; u!=NULL; u=u->next) {
				ConstructACLMessage(u, s);
			}
			netams_rwlock_unlock(&Units->rwlock);
		}
		cfg->uptime=c_uptime+1;

		// 2. check the input queue
		msgs_proc=msgs_proc_ok=0;
		while((msg=cfg->in->TryPop())) {
			if (msg->type==MSG_ACLSERVER) {
				msgacl=(Message_AclServer*)msg;
				inet_ntop(AF_INET, &(msgacl->ip), ip_s, 32);
				aDebug(DEBUG_ACLSERVER, "message ip=%s action=%s\n", ip_s, msgacl->flag?"ADD":"REMOVE");
				// 2.1. send the request 
				if (msgacl->flag) { // add
					if (cfg->direction) { // add-dst
						//access-template alc-num dyn-acl-name any host XXX 
						snprintf(buffer_tx, 256, "access-template %u %s any host %s\n", cfg->acl_number, cfg->dynamic_name?cfg->dynamic_name:"", ip_s);
					} else { // add-src
						//access-template alc-num dyn-acl-name any host XXX 
						snprintf(buffer_tx, 256, "access-template %u %s host %s any\n", cfg->acl_number, cfg->dynamic_name?cfg->dynamic_name:"", ip_s);
					}
				} 
				else { //remove
					if (cfg->direction) { // remove-dst
						//access-template alc-num dyn-acl-name any host XXX 
						snprintf(buffer_tx, 256, "clear access-template %u %s any host %s\n", cfg->acl_number, cfg->dynamic_name?cfg->dynamic_name:"", ip_s);
					} else { // remove-src
						//access-template alc-num dyn-acl-name any host XXX 
						snprintf(buffer_tx, 256, "clear access-template %u %s host %s any\n", cfg->acl_number, cfg->dynamic_name?cfg->dynamic_name:"", ip_s);
					}
				}// remove
				msgs_proc++;
				if (sAclServer_SendCommand(buffer_tx, buffer_rx, 1024*4, &fd, cfg, res)) {
					aDebug(DEBUG_ACLSERVER, "acl command \"%s\" failed \n", buffer_tx);
				} else msgs_proc_ok++;
			} else {
				aDebug(DEBUG_ACLSERVER, "message type %u not ACLSERVER - discarded\n", msg->type);
			}
			msg=cfg->in->Pop();
		}		
		
		// 99. close and clean
		if (fd) {
			shutdown(fileno(fd), SHUT_RDWR);
			fclose(fd);
		}		
		freeaddrinfo(res); 
		aDebug(DEBUG_ACLSERVER, "messages processed: %u, failed: %u\n", msgs_proc, msgs_proc-msgs_proc_ok);
	}

	pthread_cleanup_pop(1);
	return NULL;
}

//////////////////////////////////////////////////////////////////////////
void sAclServerCancel(void *v){
	Service *s = (Service*)v;
	ServiceAclServer_cfg *cfg=(ServiceAclServer_cfg*)s->cfg;
	aLog(D_INFO, "cancelling service AclServer\n");
	if (cfg->dynamic_name) aFree(cfg->dynamic_name);
	if(cfg->hostname) aFree(cfg->hostname);
	delete cfg->in;
}

//////////////////////////////////////////////////////////////////////////
void sAclServerListCfg(Service *s, FILE *f, u_char flags){
	ServiceAclServer_cfg *cfg=(ServiceAclServer_cfg*)s->cfg;

	fprintf(f, "hostname %s", cfg->hostname);
	if (cfg->port!=S_ACLSERVER_DEF_cmdport) fprintf(f, " %u", cfg->port);
	fprintf(f, "\ndirection %s\n", cfg->direction?"dst":"src");
	if (cfg->dynamic_name) fprintf(f, "dynamic-name %s\n", cfg->dynamic_name);
	fprintf(f, "acl-number %u %s\n", cfg->acl_number, cfg->is_cisco?"cisco":"");
	if (cfg->delay!=S_ACLSERVER_DEF_delay) fprintf(f, "delay %u\n", cfg->delay);
	fprintf(f, "\n"); // finish the service config
}
//////////////////////////////////////////////////////////////////////////
unsigned long long sAclServer_ParseUptime(char *buffer){
	/* What we can receive in the buffer:
	 * 
	 * dl-www:~#rsh 223.313.57.30 "show version | i uptime"
	 * cisco-rtr uptime is 1 week, 2 days, 9 hours, 39 minutes
	 * 
	 * i.e. we can get from 1 to 5 numbers here, years to minutes
	 * 
	 */
	 char *p = strchr(buffer, ' ');
	 if (p) buffer=p+1; else return 0L;
	 unsigned int value[5]; 
	 int i;
	 unsigned long long ret;
	 for (i=0; i<5; i++) value[i]=0;
	 for (i=0; buffer[i]!=0; i++) if (!isdigit(buffer[i])) buffer[i]=' '; // wipe all but numbers 
	 i=sscanf(buffer, "%u%u%u%u%u", &value[0],&value[1],&value[2],&value[3],&value[4]);
	 if (value[0]+value[1]+value[2]+value[3]+value[4]==0) return 0L;
	 //printf("\t1(%u)\t", i); for (int k=0; k<5; k++) printf("%u ", value[k]); printf("\n");
	 memmove(&value[5-i], value, sizeof(unsigned int)*(i)); bzero(value, sizeof(unsigned int)*(5-i));
	 //printf("\t2\t"); for (int k=0; k<5; k++) printf("%u ", value[k]); printf("\n");
	 ret=value[0]*365*24*60*60 + value[1]*7*24*60*60 + value[2]*24*60*60 + value[3]*60*60 + value[4]*60;
	 return ret;
}
//////////////////////////////////////////////////////////////////////////
int sAclServer_SendCommand(char *tx, char *rx, int max_rx, FILE **fd, ServiceAclServer_cfg *cfg, struct addrinfo *res){
    int sock, lport;
	
	if (*fd == NULL) { // reopen connection

		sock=rresvport(&lport);
		if (sock==-1) {
			aDebug(DEBUG_ACLSERVER, "acl socket src bind to privileged port failed: %s \n", strerror(errno));
			return -1;
		}

		if (connect(sock, res->ai_addr, res->ai_addrlen) < 0) {
			aDebug(DEBUG_ACLSERVER, "acl connect error: %s \n", strerror(errno));
			return -2;
		}

		*fd = fdopen(sock, "w+");
		if (fd==NULL) {
			aDebug(DEBUG_ACLSERVER, "acl fdopen error: %s \n", strerror(errno));
			return -3;
		}
	}
			
	// processing starts here
	// 0. send the protocol-specific junk
	fputc('0', *fd); fputc(0, *fd); fflush(*fd);
	fprintf(*fd, "root"); fputc(0, *fd); fprintf(*fd, "netams"); fputc(0, *fd);
	
	// 1. get the uptime
	fprintf(*fd, "%s", tx); fputc(0, *fd);
	fflush(*fd);
	fgetc(*fd);
	fgets(rx, max_rx, *fd);
	
	if (cfg->is_cisco) {
		shutdown(fileno(*fd), SHUT_RDWR);
		fclose(*fd);
		*fd = NULL;
	}
	return 0;
}
//////////////////////////////////////////////////////////////////////////
void ConstructACLMessage(NetUnit *u, Service *acl) {
	struct in_addr addr;
	addr.s_addr = INADDR_ANY;
	
	if (u->type==NETUNIT_HOST || u->type==NETUNIT_USER) {
		switch(u->type) {
			case NETUNIT_HOST:
				addr.s_addr = ((NetUnit_host*)u)->ip.s_addr;
				break;
			case NETUNIT_USER:
				addr.s_addr = ((NetUnit_user*)u)->ip.s_addr;
				break;
			default:
				break;						 
		}
		if (addr.s_addr!=INADDR_ANY) {
			Message_AclServer *msg = (Message_AclServer*)MsgMgr->New(MSG_ACLSERVER);
			msg->flag = u->sys_policy?ADD:REMOVE;
			msg->ip.s_addr = addr.s_addr;
			ServiceAclServer_cfg *cfg=(ServiceAclServer_cfg*)acl->cfg;
			cfg->in->Push(msg);
			aDebug(DEBUG_ACLSERVER, "queue u=%06X flag=%u sp_now=%u\n", u->id, msg->flag, u->sys_policy);
		} 
	}
}
//////////////////////////////////////////////////////////////////////////
