/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: s_datasource.c,v 1.165.2.9 2006/03/24 18:04:56 anton Exp $ */

#include "netams.h"

void *sDataSource(void *s);
void sDSCancel(void *);
void sDS_RuleIp(u_short flag, Service *s, unsigned short number, char *rule); // flag==0 eq ADD, ==1 eq REMOVE
void AcctFlows(ServiceDS_cfg* cfg);
void NFSourceAdd(ServiceDS_cfg *cfg, char *param, u_char no_flag);
//////////////////////////////////////////////////////////////////////////
//ds's we support
void* ds_ipq(void *ss);
void* ds_ulog(void *ss);
void* ds_ipfw(void *ss);
void* ds_netgraph(void *ss);
void* ds_netflow(void *ss);
void* ds_libpcap(void *ss);
void* ds_mail(void *ss);

extern void ds_libpcap_stats(FILE *f, ServiceDS_cfg *cfg);
extern void ds_netgraph_stats(FILE *f, ServiceDS_cfg *cfg);
//////////////////////////////////////////////////////////////////////////
void sDSInit(Service *s){
	s->cfg = (ServiceDS_cfg*)aMalloc(sizeof(ServiceDS_cfg));

	ServiceDS_cfg *cfg=(ServiceDS_cfg*)s->cfg;
	cfg->type=PT_UNKNOWN;
	
	cfg->id=s->instance;
	cfg->flags=DS_NONE;

	cfg->total_packets=0;
	cfg->total_flows=0;
	cfg->total_errors=0;
	cfg->timestamp=0;

	cfg->root=NULL;
	cfg->src_cmd=NULL;
	
	//netflow
	cfg->l_addr.s_addr=INADDR_ANY;
	cfg->port=0;
	cfg->nfroot=NULL;

	//libpcap
	cfg->pcap_data=NULL;
	
	//ulog
	cfg->nlmask=0;

	cfg->delay=0; cfg->skewdelay=0;
	cfg->pc=cfg->bc=cfg->tc=0;

	cfg->IPtree = new IPTree(s->instance);
	cfg->FE = NULL;
	cfg->msg = NULL;
        cfg->flow = cfg->ds_flow = NULL;
	cfg->ds_fifo = NULL;

	cfg->max_flow_slots = 1;

	cfg->packet=(unsigned char*)aMalloc(MAX_PKT_SIZE);

	cfg->child=NULL;

	cfg->layer7_detect=LAYER7_DETECT_NONE;
	
	pthread_create(&(s->t_id), NULL, &sDataSource, s);
}

//////////////////////////////////////////////////////////////////////////
void sDSProcessCfg(char *param[], Connection *conn, u_char no_flag){
	ServiceDS_cfg *cfg=(ServiceDS_cfg*)conn->service->cfg;
	char buffer[32];

	if (!strcasecmp(param[0], "type")) {
		if (!strcasecmp(param[1], "ip-traffic")) {
			aParse(conn, "data-source type is set to ip-traffic\n");
			cfg->type=PT_IP_TRAFFIC;
			cfg->max_flow_slots=3;
#ifdef LAYER7_FILTER
			cfg->max_flow_slots++;
#endif
		}
#ifdef FREEBSD
		else if (!strcasecmp(param[1], "netgraph")) {
			aParse(conn, "data-source type is set to netgraph\n");
			cfg->type=PT_NETGRAPH;
			cfg->max_flow_slots=4;
		}
#endif		
		else if (!strcasecmp(param[1], "netflow")) {
			aParse(conn, "data-source type is set to cisco NetFlow\n");
			cfg->type=PT_NETFLOW_TRAFFIC;
			if (!cfg->port) cfg->port=20001;
			if(cfg->src_cmd) aFree(cfg->src_cmd);
			cfg->src_cmd=set_string("0.0.0.0");
			cfg->max_flow_slots=5;
		}
#ifndef NOPCAP
		else if (!strcasecmp(param[1], "libpcap")) {
			aParse(conn, "data-source type is set to libpcap\n");
			cfg->type=PT_LIBPCAP_TRAFFIC;
			cfg->max_flow_slots=4;
#ifdef LAYER7_FILTER
			cfg->max_flow_slots++;
#endif
		}
#endif
		else aParse(conn, "data-source type is invalid\n");
	}
	else if (!strcasecmp(param[0], "rule")) {
		unsigned number=strtol(param[1], NULL, 10);
		ds_rule *t=NULL, *tt=NULL;

		for (t=cfg->root; t!=NULL; t=t->next) {
			if (t->number==number) break;
			tt=t;
		}

		if (no_flag && t) {
			aParse(conn, "rule %d removed\n", number);
			sDS_RuleIp(REMOVE, conn->service, number, t->rule_str);
			if( t->rule_str) aFree(t->rule_str);
			if(cfg->root==t) cfg->root=t->next;
			else tt->next=t->next;
			aFree(t);
		} //chain removal!!!!!!!
		else if (!no_flag && t) {
			sDS_RuleIp(REMOVE, conn->service, number, t->rule_str);
			if(t->rule_str) aFree(t->rule_str);
			t->rule_str=set_string(param[2]);
			aParse(conn, "rule %d replaced to %s\n", number, t->rule_str);
			sDS_RuleIp(ADD, conn->service, number, t->rule_str);
		}
		else if (!t) {
			t=(ds_rule*)aMalloc(sizeof(ds_rule));
			t->rule_str=set_string(param[2]);
			t->number=number;
			t->next=NULL;
			if(cfg->root) tt->next=t; else cfg->root=t;
			aParse(conn, "rule %d set to %s\n", number, t->rule_str);
			sDS_RuleIp(ADD, conn->service, number, t->rule_str);
		}
	}
	else if (!strcasecmp(param[0], "source")) {
		if (0) { } // mailstat was here
#if defined(LINUX)
		else if (!strcasecmp(param[1], "ipq")) {
			cfg->src_cmd=set_string("ipq");
			cfg->flags=DS_IPQ;
			aParse(conn, "source is set to ipq\n");
		}
		else if (!strcasecmp(param[1], "ulog")) {
			unsigned nlmask=0;
			u_char i=2;
			while(param[i]!=empty && isdigit(param[i][0])) {
				u_char nlgroup=strtol(param[i], NULL, 10);
				if(nlgroup<1 || nlgroup>32)
					aParse(conn, "nlgroup has to be between 1 and 32\n");
				else 
					nlmask|=(1<<(nlgroup-1));
				
				i++;
			}
			if(nlmask && nlmask!=cfg->nlmask) {
				if(cfg->src_cmd) aFree(cfg->src_cmd);
				cfg->src_cmd=set_string("ulog");
				cfg->nlmask=nlmask;
				cfg->flags=DS_ULOG;
				aParse(conn, "source is set to ulog, nlmask=%u\n", cfg->nlmask);
			}
        	}
#endif
#if defined(FREEBSD)
		else if (!strcasecmp(param[1], "tee")) {
			unsigned short port=strtol(param[2], NULL, 10);
			if (!port) port=199;
			aParse(conn, "source is set to tee:%u\n", port);
			cfg->src_cmd=set_string("tee");
			cfg->port=port;
			cfg->flags=DS_TEE;
		}
		else if (!strcasecmp(param[1], "divert")) {
			unsigned short port=strtol(param[2], NULL,10);
			if (!port) port=199;
			aParse(conn, "source is set to divert:%u\n", port);
			cfg->src_cmd=set_string("divert");
			cfg->port=port;
			cfg->flags=DS_DIVERT;
		}
#endif
		else if (cfg->type == PT_NETFLOW_TRAFFIC) {
			if(INADDR_BROADCAST!=inet_addr(param[1])) {
				NFSourceAdd(cfg, param[1], no_flag);
				aParse(conn, "%s %s source %s accept list\n", (no_flag)?"Removing":"Adding", param[1],(no_flag)?"from":"to");
			}
		}
		else if (cfg->type==PT_LIBPCAP_TRAFFIC) {
			aParse(conn, "source is listening interface %s\n", param[1]);
			cfg->src_cmd=set_string(param[1]);
			cfg->port=0;
			if(!strcasecmp(param[2], "promisc")) cfg->flags=DS_PROMISC;
			else cfg->flags=DS_NONE;
		}
#ifdef FREEBSD
		else if (cfg->type==PT_NETGRAPH) {
			aParse(conn, "source is listening node %s\n", param[1]);
			cfg->src_cmd=set_string(param[1]);
			cfg->port=0;
			if(!strcasecmp(param[2], "divert")) cfg->flags=DS_DIVERT;
			else cfg->flags=DS_TEE;
		}
#endif
		else aParse(conn, "source is invalid or not available\n");
	}
	else if (!strcasecmp(param[0], "clock")) {
		if(cfg->type == PT_NETFLOW_TRAFFIC) {
			if (!strcasecmp(param[1], "remote")) {
				aParse(conn, "using remote flow time\n");
				cfg->flags=DS_CLOCK_REMOTE;
			} else if (!strcasecmp(param[1], "local")) {
				aParse(conn, "using local time\n");
				cfg->flags=DS_NONE;
			} else {
				aParse(conn, "Clock undefined! Defaulting to local time\n");
				cfg->flags=DS_NONE;
			}
		} else {
			aParse(conn, "Setting clock undefined for this data-source type\n");
		}
	}
	else if (!strcasecmp(param[0], "listen")) {
		if (cfg->type == PT_NETFLOW_TRAFFIC) {
			if(INADDR_BROADCAST!=inet_addr(param[1])) {
				inet_aton(param[1], &cfg->l_addr);
				if(param[2]!=empty) cfg->port=strtol(param[2], NULL, 10);
				if(cfg->src_cmd) aFree(cfg->src_cmd);
				cfg->src_cmd=set_string(param[1]);
				aParse(conn, "listening %s:%u\n", inet_ntop(AF_INET, &(cfg->l_addr), buffer, 32), cfg->port);
			} else {
				aParse(conn, "Command listen undefined for this data-source type\n");
                	}
		}
	}
#ifdef LAYER7_FILTER
	else if (!strcasecmp(param[0], "layer7-detect")) {
		if (cfg->type == PT_LIBPCAP_TRAFFIC || cfg->type == PT_IP_TRAFFIC) {
			if (!strcasecmp(param[1], "none")) {
				aParse(conn, "do not detect any type of layer7 traffic\n");
				cfg->layer7_detect=LAYER7_DETECT_NONE;
			} else if (!strcasecmp(param[1], "urls")) {
				aParse(conn, "detect URLs\n");
				cfg->layer7_detect=LAYER7_DETECT_URLS;
			} else aParse(conn, "layer7-detect parameter unknown\n");
		} else aParse(conn, "layer7-detect will not work for this data-source type\n");
	}
#endif
	else aParse(conn, "unknown data-source command: %s\n", param[0]);
}
//////////////////////////////////////////////////////////////////////////
void *sDataSource(void *ss){
	Service *s=(Service*)ss;

	aLog(D_INFO, "service data-source:%u thread started (%p)\n", s->instance, s->cfg);
	pthread_cleanup_push(sDSCancel, s);
	SET_CANCEL();
	s->t_id=pthread_self();

	ServiceDS_cfg *cfg=(ServiceDS_cfg*)s->cfg;

	// now we should sleep before starting data source service exec
	if(!(s->flags&SERVICE_RUN)) s->Sleep();
	
	//init units for this datasource
        //organize tree here
	Units->units2ds(s, ADD);
	
	u_char		fw_check 		= 0;
	void		*(*ds_func)(void*)	= NULL;
	service_type	ds_type			= SERVICE_UNDEF;

	switch (cfg->type) {
		case PT_IP_TRAFFIC:	// here we are assume that arrived data has an IP format
			#if defined(LINUX) && !defined(IPTBL_NONE)
			if(cfg->flags==DS_IPQ) {
				fw_check = 1;
				ds_type	 = SERVICE_DS_IPQ;	 
				ds_func	 = &ds_ipq; 		
			} else {
				fw_check = 0;
				ds_type  = SERVICE_DS_ULOG;
				ds_func  = &ds_ulog;
			}
			#endif
			#ifdef FREEBSD
				fw_check = (cfg->flags==DS_DIVERT);
				ds_type  = SERVICE_DS_IPFW;
				ds_func  = &ds_ipfw;
			#endif
			break;
		case PT_NETFLOW_TRAFFIC:
			ds_type  = SERVICE_DS_NETFLOW;
			ds_func  = &ds_netflow;
			break;
#ifndef NOPCAP
		case PT_LIBPCAP_TRAFFIC:
			ds_type  = SERVICE_DS_LIBPCAP;
			ds_func  = &ds_libpcap;
			break;
#endif
#ifdef FREEBSD
		case PT_NETGRAPH: 
			ds_type  = SERVICE_DS_NETGRAPH;
			ds_func  = &ds_netgraph;
			break;
#endif
		default:	// this source type is not catched yet!
			aLog(D_ERR, "Data-source:%u type is undefined or not compiled\n", s->instance);
			goto END;
			break;
	}

	cfg->ds_fifo = new DS_FIFO(10000);
	cfg->FE = new FlowEngine(s, fw_check);
	
	cfg->child=new Service(ds_type, s->instance);
	pthread_create(&(cfg->child->t_id), NULL, ds_func, ss);
	Services->Insert(cfg->child);
	
	while(1) {
		AcctFlows(cfg);
		s->Sleep(1);
	}

END:
	pthread_cleanup_pop(1);
	        
	//cleanup if service can't be started
	Services->Delete(s);
	delete s;

	return NULL;
}
//////////////////////////////////////////////////////////////////////////
void sDSCancel(void *v){
	Service *s = (Service*)v;
	ServiceDS_cfg *cfg=(ServiceDS_cfg*)s->cfg;
	
	ds_rule *ptr,*t=cfg->root;
	while(t) {
//		printf("removing %p (%p) %d %s\n", t, t->next, t->number, t->rule_str);
		sDS_RuleIp(REMOVE, s, t->number, t->rule_str);
		if(t->rule_str) aFree(t->rule_str);
		ptr=t; t=t->next;
		aFree(ptr);
	}
	if(cfg->child) {
		cfg->child->Shutdown();
		Services->Delete(cfg->child);
		delete cfg->child;
	}

	if(cfg->FE) delete cfg->FE;
	if(cfg->ds_fifo) {
		aLog(D_INFO, "cancelling service data-source:%u, flushing %u messages\n", s->instance, cfg->ds_fifo->num_items);
		s->flags|=SERVICE_RUN;  //this is workaround for auto-units
		AcctFlows(cfg);  //ensure we count collected data
		s->flags&=~SERVICE_RUN; //this is workaround for auto-units

		delete cfg->ds_fifo;
	}
	if(cfg->flow) delete cfg->flow;
	if(cfg->ds_flow) delete cfg->ds_flow;

	delete cfg->IPtree;
	
	for(NFSource *tmp = cfg->nfroot; cfg->nfroot!=NULL; cfg->nfroot = tmp) {
		tmp=cfg->nfroot->next;
		aFree(cfg->nfroot);
	}

	if(cfg->pcap_data) aFree(cfg->pcap_data);
	if(cfg->packet) aFree(cfg->packet);
	if(cfg->src_cmd) aFree(cfg->src_cmd);
}

//////////////////////////////////////////////////////////////////////////
void sDSListCfg(Service *s, FILE *f){
	ServiceDS_cfg *cfg=(ServiceDS_cfg*)s->cfg;
	char buffer[32];

	if (cfg->type==PT_IP_TRAFFIC){
		fprintf(f, "type ip-traffic\n");
		
		fprintf(f, "source %s", cfg->src_cmd);

		if(cfg->flags==DS_ULOG) {
			for(u_char i=0;i<32;i++)
				if(cfg->nlmask&(1<<i)) fprintf(f, " %u", i+1);
		} 
		else if (cfg->flags==DS_TEE || cfg->flags==DS_DIVERT)
			fprintf(f, " %u", cfg->port);
		
		fprintf(f, "\n");
	}
	else if (cfg->type==PT_NETFLOW_TRAFFIC) {
		in_addr ina;

		fprintf(f, "type netflow\n");
		if (cfg->l_addr.s_addr != INADDR_ANY || cfg->port!=20001) 
			fprintf(f, "listen %s %u\n", inet_ntop(AF_INET, &(cfg->l_addr), buffer, 32), cfg->port);
		for(NFSource *nf=cfg->nfroot; nf!=NULL; nf=nf->next) {
			ina.s_addr=nf->src_addr;
			fprintf(f, "source %s\n", inet_ntop(AF_INET, &ina, buffer, 32));
		}
		if (cfg->flags==DS_CLOCK_REMOTE) fprintf(f, "clock remote\n");
		fprintf(f, "\n");
		return;
	}
	else if (cfg->type==PT_LIBPCAP_TRAFFIC){
		fprintf(f, "type libpcap\n");
		fprintf(f, "source %s %s\n", cfg->src_cmd, (cfg->flags==DS_PROMISC)?"promisc":"");
	}
	else if (cfg->type==PT_NETGRAPH){
		fprintf(f, "type netgraph\n");
		fprintf(f, "source %s %s\n", cfg->src_cmd, (cfg->flags==DS_DIVERT)?"divert":"");
	}
	for (ds_rule *r=cfg->root; r!=NULL; r=r->next) 
		if (r->rule_str) fprintf(f, "rule %u \"%s\"\n", r->number, r->rule_str);

	if (cfg->layer7_detect!=LAYER7_DETECT_NONE){
		if (cfg->layer7_detect==LAYER7_DETECT_URLS) fprintf(f, "layer7-detect urls\n");
	}

	fprintf(f, "\n");
}
//////////////////////////////////////////////////////////////////////////
void sDS_RuleIp(u_short flag, Service *s, u_short number, char *rule) {
	ServiceDS_cfg *cfg=(ServiceDS_cfg*)s->cfg;
	
	if (cfg->type==PT_IP_TRAFFIC) {
		char buffer[255];
		
		switch (flag) {
		case ADD: // add a rule to the system
			#ifdef LINUX
			sprintf(buffer,"/sbin/iptables -I %s 2>&1 >/dev/null", rule);
			#else
			sprintf(buffer,"/sbin/ipfw add %u %s %u %s 2>&1 >/dev/null", number, cfg->src_cmd, cfg->port, rule);
			#endif
			
			break;
		case REMOVE:  // remove a rule from the system
			#ifdef LINUX
			sprintf(buffer, "/sbin/iptables -D %s 2>&1 >/dev/null", rule);
			#else
			sprintf(buffer,"/sbin/ipfw delete %u 2>&1 >/dev/null", number);
			#endif
	
			break;
		default: 
			break;
		}
		system(buffer);
	}
	aLog(D_INFO, "%s rule %u \"%s\"\n", (flag==ADD)?"adding":"removing", number, rule);
}
//////////////////////////////////////////////////////////////////////////
void NFSourceAdd(ServiceDS_cfg *cfg, char *param, u_char no_flag) {
	NFSource *nf, *p=NULL;
	in_addr ina;
	
	inet_aton(param, &ina);

	for(nf=cfg->nfroot; nf!=NULL; nf=nf->next) {
		if(nf->src_addr==ina.s_addr) break;
		p=nf;
	}

	if(no_flag) { //REMOVE
		if(!nf) return;
		if(cfg->nfroot==nf) cfg->nfroot=cfg->nfroot->next;
		else p->next=nf->next;
		if(nf->nfTemplate.fields != NULL)
		    aFree(nf->nfTemplate.fields);
		aFree(nf);
	} else { //ADD
		if(nf) return;
		nf=(NFSource*)aMalloc(sizeof(NFSource));
		nf->src_addr=ina.s_addr;
		if(cfg->nfroot==NULL) cfg->nfroot=nf;
		else p->next=nf;
	}
}
//////////////////////////////////////////////////////////////////////////
void cShowDS(Connection *conn){
        Service *s=NULL;
	char buffer[32];

        while((s=Services->getServiceNextByType(SERVICE_DATASOURCE,s))) {
		ServiceDS_cfg *cfg=(ServiceDS_cfg*)s->cfg;
		
		sDSShow(s, conn->stream_w);

		//show IP tree
		IPTree *IPtree=cfg->IPtree;
		fprintf(conn->stream_w, "    IP tree: %lu nodes [%u] + %lu dlinks [%u] + %lu unodes [%u] = %lu bytes\n",IPtree->nodes,sizeof(IPTree_node),IPtree->dlink,256*sizeof(IPTree_node*),IPtree->unodes,sizeof(dsUlist),IPtree->nodes*sizeof(IPTree_node)+IPtree->dlink*256*sizeof(IPTree_node*)+IPtree->unodes*sizeof(dsUlist));
		//show flow
		FlowEngine *FE=cfg->FE;
       		if(FE &&
			//this actually doesnt use FlowEnginei completely, only DoSend() call.
			!(cfg->type==PT_NETFLOW_TRAFFIC	|| cfg->type==PT_NETGRAPH)
		) {
			fprintf(conn->stream_w, "    Flows: %u/%u act/inact entries (%u bytes), %llu flows sent\n",
				FE->e_used,
				FE->e_total-FE->e_used,
				FE->e_total*(sizeof(entry)+sizeof(Flow)+sizeof(struct attribute)*cfg->max_flow_slots),
				FE->sent_flows);
			//show hash
			//gather flow hash statistic
        		entry *e;
			entry **table=FE->table;
        		unsigned used=0;
        		unsigned max_chain=0;
        		unsigned tmp;
        		unsigned long total=0;

			for(unsigned i=0;i<FLOW_HASH_SIZE;i++) {
                		if(!(e=table[i])) continue;
                		used++;
                		tmp=0;
				for(;e!=NULL;e=e->next) {
					tmp++;
				}
				total+=tmp;
				if(max_chain<tmp) max_chain=tmp;
        		}

			fprintf(conn->stream_w, "    HASH: size=%u, %lu flows hashed, %u nodes used, max chain= %u\n",FLOW_HASH_SIZE,total,used,max_chain);
#ifdef HAVE_BW			
			BWEngine *bw = FE->BW;
			if(bw) {
				BWdata *b;
				BWdata **table=bw->table;
				unsigned used=0;
				unsigned max_chain=0;
				unsigned tmp;
				unsigned long total=0;

				for(unsigned i=0;i<BW_HASH_SIZE;i++) {
					if(!(b=table[i])) continue;
					used++;
					tmp=0;
					for(;b!=NULL;b=b->next) {
						tmp++;
					}
					total+=tmp;
					if(max_chain<tmp) max_chain=tmp;
				}

				fprintf(conn->stream_w, "    BW HASH: size=%u, %lu bws hashed , %u nodes used, max chain= %u\n",
					BW_HASH_SIZE, total, used, max_chain);
					
				fprintf(conn->stream_w, "    BW HASH: %lu/%lu slots/ready, %lu/%lu wrappers/ready, %u bytes\n",
					bw->bwd_used, bw->bwd_total-bw->bwd_used,
					bw->bwu_used, bw->bwu_total-bw->bwu_used,
					bw->bwd_total*sizeof(BWdata)+bw->bwu_total*sizeof(bwUlist)
				);
			}
#endif
		}
		
		//show fifo
		DS_FIFO *fifo=cfg->ds_fifo;
		unsigned msg_size=sizeof(EngineMsg)+sizeof(struct attribute)*cfg->max_flow_slots;
		if(fifo) fprintf(conn->stream_w, "    FIFO: %u/%u used/ready messages, each %u, total %u bytes\n", fifo->num_items,fifo->ready_items, msg_size,(fifo->num_items+fifo->ready_items)*msg_size);
		
		//show sources info
		if(cfg->type==PT_NETFLOW_TRAFFIC) {
			char buf[32];
			in_addr ina;
			fprintf(conn->stream_w, "    Total: %llu packets, %llu flows, %u errors, timestamp - %s\n", cfg->total_packets, cfg->total_flows, cfg->total_errors, (cfg->timestamp)?timeU2T((time_t)cfg->timestamp, buf):"never");
			for(NFSource *nf=cfg->nfroot; nf!=NULL; nf=nf->next) {
				ina.s_addr=nf->src_addr;
				fprintf(conn->stream_w, "    Source %s: %llu packets, %llu flows, %u errors, timestamp - %s\n", inet_ntop(AF_INET, &ina, buffer, 32), nf->packets, nf->flows, nf->errors, (nf->timestamp)?timeU2T((time_t)nf->timestamp, buf):"never");
			}
#ifndef NOPCAP
		} else if (cfg->type==PT_LIBPCAP_TRAFFIC) {
			ds_libpcap_stats(conn->stream_w, cfg);
#endif
#ifdef FREEBSD
		} else if (cfg->type==PT_NETGRAPH) {
			ds_netgraph_stats(conn->stream_w, cfg);
#endif
		}
	}
}
//////////////////////////////////////////////////////////////////////////
void AcctFlows(ServiceDS_cfg* cfg) {
	DS_FIFO *fifo	= cfg->ds_fifo;
	IPTree *IPtree	= cfg->IPtree;
	EngineMsg *msg;
	Flow *flow	= cfg->flow;
	struct attribute *tmp;

	while((cfg->msg=fifo->Pop(cfg->msg))) {
 		msg=cfg->msg;
		flow->freeSlot= msg->freeSlot;
		//swap lists
		tmp=msg->list;
		msg->list=flow->list;
		flow->list=tmp;             	
		
		IPtree->ACCT(flow);
#ifdef LAYER7_FILTER
		layer7_freeinfo(flow);
#endif
	}
}
//////////////////////////////////////////////////////////////////////////
unsigned sDSMeasure(ServiceDS_cfg* cfg, struct timeval *start, unsigned len) {
	unsigned diff;
	struct timeval stop;

	netams_gettimeofday(&stop, NULL);
	
	diff= (unsigned) ( (stop.tv_sec - start->tv_sec)*1000000 + (stop.tv_usec - start->tv_usec));
	
	cfg->total_packets++;
	cfg->delay += diff;
	cfg->skewdelay = (unsigned)(cfg->skewdelay * 98 + diff * 2)/100;
	
	if(start->tv_sec - cfg->tc > LOAD_INTERVAL) {
		cfg->ptc = cfg->tc;	cfg->tc	= start->tv_sec;
		cfg->ppc = cfg->pc;	cfg->pc	= 0;
		cfg->pbc = cfg->bc;	cfg->bc = 0;
	}

	cfg->pc++;
	cfg->bc += len;

	aDebug(DEBUG_DS_IP, "ds:%u packet processed in %u mcsec\n",cfg->id, diff);

	return diff;
}
//////////////////////////////////////////////////////////////////////////
void sDSShow(Service *s, FILE *f) {
	ServiceDS_cfg *cfg=(ServiceDS_cfg*)s->cfg;

	unsigned avg_pps;
	u_long avg_bps;

	if(cfg->tc - cfg->ptc <= 2*LOAD_INTERVAL) {
		avg_pps = cfg->ppc/(LOAD_INTERVAL);
		avg_bps = cfg->pbc/(LOAD_INTERVAL);
	} else {
		avg_pps = 0;
		avg_bps = 0;
	}
	
	char *pt;
                
	switch (cfg->type){
		case PT_IP_TRAFFIC: pt="IP_FILT"; break;
		case PT_NETFLOW_TRAFFIC: pt="NETFLOW"; break;
		case PT_LIBPCAP_TRAFFIC: pt="LIBPCAP"; break;
		case PT_NETGRAPH: pt="NETGRAPH"; break;
		default : pt="<\?\?>"; break;
	}

	fprintf(f, " Data-source ID=%d type %s source %s:%u loop %llu average %d mcsec\n",
		s->instance, pt, cfg->src_cmd?cfg->src_cmd:"<>", cfg->port,
		cfg->total_packets, (cfg->total_packets)?(unsigned)(cfg->delay/cfg->total_packets):0
	);
                
	fprintf(f, "    Perf: average skew delay %u mcsec, PPS: %u, BPS: %lu\n",
		cfg->skewdelay, avg_pps, avg_bps);
}
//////////////////////////////////////////////////////////////////////////
