/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: s_login.c,v 1.85.2.6 2006/03/24 18:04:55 anton Exp $ */

#include "netams.h"

Service *Login=NULL;

void *sLogin(void *s);
void sLgCancel(void *);
void sLgProcessData(Login_cfg *cfg);
void sLgObtainDbData(NetUnit *u, Login_cfg *c);
void sLgSetCfg(char *param[], Connection *conn, Login_cfg *c);
void sLgLoginCfg(char *param[], Connection *conn, u_char no_flag, Login_cfg *c);
void sLgObtainMac(struct in_addr *u, struct ether_addr *current_mac);
void sLgRestoreLogin(Login_cfg *cfg);

#define LG_LOGIN	0x01
#define LG_LOGOUT	0x02
#define LG_FORCE	0x04
#define LG_SET_USER_IP	0x08

void sLgLogin(Connection *conn, NetUnit *u, struct in_addr *ip, Login_cfg *cfg, u_char flag);
//////////////////////////////////////////////////////////////////////////////////////////
void sLgInit(Service *s){
	Login_cfg *cfg = (Login_cfg*)aMalloc(sizeof(Login_cfg));
	cfg->default_inact=S_LOGIN_DEF_default_inact;
	cfg->default_abs=S_LOGIN_DEF_default_abs;
	cfg->max_inact=S_LOGIN_DEF_max_inact;
	cfg->min_inact=S_LOGIN_DEF_min_inact;
	cfg->max_abs=S_LOGIN_DEF_max_abs;
	cfg->min_abs=S_LOGIN_DEF_min_abs;
	cfg->min_passwd_length=S_LOGIN_DEF_min_passwd_length;
	cfg->storage=0;
	cfg->set_user_ip=0;
	cfg->st=NULL;
	cfg->relogin=1;
	
	print_to_string(&cfg->file, "login.%u", s->instance);

        netams_rwlock_init(&cfg->rwlock, NULL);

	s->cfg=cfg;
	Login=s;
	pthread_create(&(s->t_id), NULL, &sLogin, s);
}
//////////////////////////////////////////////////////////////////////////////////////////
void sLgProcessCfg(char *param[], Connection *conn, u_char no_flag){

	Service *s=NULL;
	if(!(s=Login)) return;

	Login_cfg *cfg=(Login_cfg*)s->cfg;
	time_t t;

	netams_rwlock_wrlock(&cfg->rwlock);

	if (!strcasecmp(param[0], "")) {
	}
	else if (!strcasecmp(param[0], "default-inact")) {
		t=strtol(param[1], NULL, 10);
		if (t<0 || t>60*60*24*7) return; // disallow negative and more than one week
		cfg->default_inact=t;
		aParse(conn, "default inactivity timeout set to %ld seconds\n", t);
	}
	else if (!strcasecmp(param[0], "default-abs")) {
		t=strtol(param[1], NULL, 10);
		if (t<0 || t>60*60*24*7) return; // disallow negative and more than one week
		cfg->default_abs=t;
		aParse(conn, "default absolute timeout set to %ld seconds\n", t);
	}
	else if (!strcasecmp(param[0], "max-inact")) {
		t=strtol(param[1], NULL, 10);
		if (t<0 || t>60*60*24*7) return; // disallow negative and more than one week
		cfg->max_inact=t;
		aParse(conn, "maximum inactivity timeout set to %ld seconds\n", t);
	}
	else if (!strcasecmp(param[0], "min-inact")) {
		t=strtol(param[1], NULL, 10);
		if (t<0 || t>60*60*24*7) return; // disallow negative and more than one week
		cfg->min_inact=t;
		aParse(conn, "minimum inactivity timeout set to %ld seconds\n", t);
	}
	else if (!strcasecmp(param[0], "max-abs")) {
		t=strtol(param[1], NULL, 10);
		if (t<0 || t>60*60*24*7) return; // disallow negative and more than one week
		cfg->max_abs=t;
		aParse(conn, "maximum absolute timeout set to %ld seconds\n", t);
	}
	else if (!strcasecmp(param[0], "min-abs")) {
		t=strtol(param[1], NULL, 10);
		if (t<0 || t>60*60*24*7) return; // disallow negative and more than one week
		cfg->min_abs=t;
		aParse(conn, "minimum absolute timeout set to %ld seconds\n", t);
	}
	else if (!strcasecmp(param[0], "min-passwd-length")) {
		t=strtol(param[1], NULL, 10);
		if (t<0 || t>60*60*24*7) return; // disallow negative and more than one week
		cfg->min_passwd_length=t;
		aParse(conn, "minimum password length set to %ld symbols\n", t);
	}
	else if (!strcasecmp(param[0], "set-user-ip")) {
		if (!strcasecmp(param[1], "yes") || strtol(param[1], NULL, 10)==1) {
			aParse(conn, "unit type user ip address will be set to current ip\n");
			cfg->set_user_ip=1;
		}
		else {
			aParse(conn, "unit type user ip address will be NOT set\n");
			cfg->set_user_ip=0;
		}
	}
	else if (!strcasecmp(param[0], "relogin")) {
		if (!strcasecmp(param[1], "no")) {
			cfg->relogin=0;
			aParse(conn, "relogin parameter set to \"no\": disallow relogins\n");
		}
		else if (!strcasecmp(param[1], "yes")) {
			cfg->relogin=1;
			aParse(conn, "relogin parameter set to \"yes\": allow relogins\n");
		}
		else aParse(conn, "unknown relogin value\n");
	}
	else if (!strcasecmp(param[0], "set")) sLgSetCfg(param, conn, cfg);
	else if (!strcasecmp(param[0], "login")) sLgLoginCfg(param, conn, no_flag, cfg);
	else if (!strcasecmp(param[0], "logout")) sLgLoginCfg(param, conn, 1, cfg);
	else aParse(conn, "unknown login command: %s\n", param[0]);

	netams_rwlock_unlock(&cfg->rwlock);	
}

//////////////////////////////////////////////////////////////////////////////////////////
void sLgListCfg(Service *s, FILE *f){

	Login_cfg *cfg=(Login_cfg*)s->cfg;
	
        if(netams_rwlock_rdlock(&cfg->rwlock)) return;

	if (cfg->default_inact!=S_LOGIN_DEF_default_inact) fprintf(f, "default-inact %ld\n", cfg->default_inact);
	if (cfg->default_abs!=S_LOGIN_DEF_default_abs) fprintf(f, "default-abs %ld\n", cfg->default_abs);
	if (cfg->max_inact!=S_LOGIN_DEF_max_inact) fprintf(f, "max-inact %ld\n", cfg->max_inact);
	if (cfg->min_inact!=S_LOGIN_DEF_min_inact) fprintf(f, "min-inact %ld\n", cfg->min_inact);
	if (cfg->max_abs!=S_LOGIN_DEF_max_abs) fprintf(f, "max-abs %ld\n", cfg->max_abs);
	if (cfg->min_abs!=S_LOGIN_DEF_min_abs) fprintf(f, "min-abs %ld\n", cfg->min_abs);
	if (cfg->min_passwd_length!=S_LOGIN_DEF_min_passwd_length) fprintf(f, "min-passwd-length %d\n", cfg->min_passwd_length);
	fprintf(f, "relogin %s\n", cfg->relogin?"yes":"no");
	if (cfg->set_user_ip) fprintf(f, "set-user-ip yes\n");

	fprintf(f, "\n");
	
	netams_rwlock_unlock(&cfg->rwlock);
}

//////////////////////////////////////////////////////////////////////////////////////////
void *sLogin(void *ss){
	Service *s=(Service*)ss;
	Login_cfg *cfg=(Login_cfg*)s->cfg;

	aLog(D_INFO, "service login:%u thread started (%p)\n", s->instance, s->cfg);
	pthread_cleanup_push(sLgCancel, s);
	SET_CANCEL();
	s->t_id=pthread_self();

	// now we should sleep before starting service
	if(!(s->flags&SERVICE_RUN)) s->Sleep();

	cfg->st = aStorageGetAccepted(ST_CONN_LOGIN);
    	if (cfg->st==NULL) { 
		aLog(D_WARN, "login service requires at least one storage to be up, skipping\n");  
		goto S_LOGIN_CLEANUP_POP;
	}
	
	stSaveSql(cfg->st,cfg->file,ST_CONN_LOGIN);
	
	while(!stLgObtainDbData(cfg)) {
		aLog(D_WARN, "Service login can't obtain data from storage:%u\n",cfg->st->instance);
		s->Sleep(10);
	}
	
	sLgRestoreLogin(cfg); //restore login status back

	aLog(D_DEBUG, "service login:%u checking every processor delay: %d seconds\n", s->instance, ProcessorCfg.delay);

	while (1) {
		s->Sleep(0); // login timeouts will be checked every cfg->delay seconds
		aDebug(DEBUG_LOGIN, "Checking logins (every %d seconds)\n", ProcessorCfg.delay);
		sLgProcessData(cfg);	
	}
S_LOGIN_CLEANUP_POP:
	pthread_cleanup_pop(0);
	return NULL;
}
//////////////////////////////////////////////////////////////////////////////////////////

void sLgRestoreLogin(Login_cfg *cfg) {
	sLoginData *ld;
	NetUnit *u;

	netams_rwlock_rdlock(&Units->rwlock);
	for (u=Units->root; u!=NULL; u=u->next) {                     
		ld=u->logindata;
		if(!ld) continue;
   
		if(ld->flags&LOGIN_ACTIVE) 
			sLgLogin(cInternal, u, NULL, cfg, (LG_LOGIN|LG_FORCE));
		else 
			sLgLogin(cInternal, u, NULL, cfg, (LG_LOGOUT|LG_FORCE));
		
		ld->flags&=~LOGIN_UPDATE;
		aDebug(DEBUG_LOGIN, "Unit: %06X inact %u abs %u %s %s\n", u->id, ld->inact, ld->abs, (ld->flags&LOGIN_ACTIVE)?"ACTIVE":"", (ld->flags&LOGIN_STRICT)?"STRICT":"");
	}
	netams_rwlock_unlock(&Units->rwlock);
}
//////////////////////////////////////////////////////////////////////////////////////////
void sLgProcessData(Login_cfg *cfg) {
	time_t now=time(NULL);
	NetUnit *u;
        time_t last_used;
	FILE *file=NULL;
	sLoginData *ld;
	
	netams_rwlock_rdlock(&Units->rwlock);
	for (u=Units->root; u!=NULL; u=u->next) { 
		ld=u->logindata;
		if (!ld) continue;

		if (ld->flags&LOGIN_ACTIVE) {
			if (ld->abs && (now - ld->opened) > ld->abs) {
				aDebug(DEBUG_LOGIN, "Unit %06X absolute timeout reached: now=%ld, opened=%ld\n", u->id, now, ld->opened);
				sLgLogin(cInternal, u, NULL, cfg, (LG_LOGOUT|LG_FORCE));			
			}
			if(u->ap) last_used=u->ap->LastUsed();
			else last_used=0;
			if ((ld->inact && last_used && (now - last_used) > ld->inact)
			|| (ld->inact && !last_used && (now - ld->opened) > ld->inact)) {
				aDebug(DEBUG_LOGIN, "Unit %06X inact timeout reached: now=%ld, opened=%ld, lastused=%ld\n", u->id, now, ld->opened, last_used);
				sLgLogin(cInternal, u, NULL, cfg, (LG_LOGOUT|LG_FORCE));
			}
		}
		if (ld->flags&LOGIN_UPDATE) {
			if(!file) {
				file=fopen(cfg->file,"a");
				if(!file) {
					aLog(D_ERR, "Can't create temporary file %s: %s\n", cfg->file, strerror(errno));
					continue;
				}
				setlinebuf(file);
			}
			ld->flags&=~LOGIN_UPDATE;
			fprintf (file, "%u,%s,%lu,%lu,%lu,%lu,%u,%s,%u\n", u->id, ld->password?ld->password:"", ld->inact, ld->abs, now, ld->opened, ld->ip_from.s_addr, ether_ntoa(&ld->mac_from), ld->flags);
                }
	}
	netams_rwlock_unlock(&Units->rwlock);
	
	if(file) {
		fclose(file);
		stSaveSql(cfg->st,cfg->file,ST_CONN_LOGIN);
	}
}
//////////////////////////////////////////////////////////////////////////////////////////
void sLgCancel(void *v){
	Service *s = (Service*)v;
	Login_cfg *cfg = (Login_cfg*)s->cfg;
	
	aLog(D_INFO, "cancelling service login:%u\n", s->instance);

	sLgProcessData(cfg);
	Login=NULL;
	stCloseSql(cfg->st,ST_CONN_LOGIN);

	if(cfg->file) aFree(cfg->file);
	netams_rwlock_destroy(&cfg->rwlock);
}
//////////////////////////////////////////////////////////////////////////////////////////
void cShowLogin(Connection *conn, char *param[]){
	NetUnit *u, *ut=NULL;
	int total=0, enabled=0, opened=0, closed=0;
	static char buf1[32], buf2[32];
	char buffer[32];

	u_char i=2;
	ut = aParseUnit(param, &i);

	Service *s=Login;
        if(!s) return;
	Login_cfg *cfg=(Login_cfg*)s->cfg;

	netams_rwlock_rdlock(&cfg->rwlock);
	netams_rwlock_rdlock(&Units->rwlock);
	for (u=Units->root; u!=NULL; u=u->next){
		total++;
		if (u->logindata) {
			enabled++;
			if (u->logindata->flags&LOGIN_ACTIVE) opened++; else closed++;
			if ((ut && ut==u) || !ut) {
				fprintf(conn->stream_w, "OID: %06X (%s) %s, inact:%ld, abs:%ld\n", u->id, u->name?u->name:"<\?\?>", u->logindata->flags&LOGIN_ACTIVE?"OPENED":"CLOSED", u->logindata->inact, u->logindata->abs);
				inet_ntop(AF_INET, &(u->logindata->last_ip_from), buffer, 32);
				if (u->logindata->flags&LOGIN_ACTIVE) fprintf(conn->stream_w, " OP:%s, LU:%s, %s[%s]\n", timeU2T(u->logindata->opened, buf1), u->ap?timeU2T(u->ap->LastUsed(), buf2):"err", buffer, ether_ntoa(&u->logindata->last_mac_from));
			}
		}
	}
	netams_rwlock_unlock(&Units->rwlock);
	netams_rwlock_unlock(&cfg->rwlock);
	fprintf(conn->stream_w, "Total units: %d, enabled: %d, opened: %d, closed:%d\n", total, enabled, opened, closed);
}

//////////////////////////////////////////////////////////////////////////////////////////
void sLgSetCfg(char *param[], Connection *conn, Login_cfg *c){
	NetUnit *u;
	time_t t_inact, t_abs;
	sLoginData *ld;
	
	u_char i=1;
	u=aParseUnit(param, &i);
	if(u==NULL) { aParse(conn, "unit not exist\n"); return; }
	
	ld=u->logindata;
	if (!ld) {
		ld = (sLoginData*)aMalloc(sizeof(sLoginData));
		ld->inact=c->default_inact;
		ld->abs=c->default_abs;
		ld->opened=0L;
		bzero(&ld->ip_from, sizeof (struct in_addr));
		bzero(&ld->mac_from, sizeof (struct ether_addr));
		ld->flags=0;
		ld->password=NULL;
		u->logindata=ld;
	}

	while (param[i]!=empty){
		if (!strcasecmp(param[i], "inact")){
			t_inact=strtol(param[i+1], NULL, 10);
			if (t_inact != 0 && (c->min_inact > t_inact || t_inact > c->max_inact)) aParse(conn, "invalid inact value range, setting to default\n");
			else ld->inact=t_inact;
		}
		else if (!strcasecmp(param[i], "abs")){
			t_abs=strtol(param[i+1], NULL, 10);
			if (t_abs != 0 && (c->min_abs > t_abs || t_abs > c->max_abs)) aParse(conn, "invalid abs value range, setting to default\n");
			else ld->abs=t_abs;
		}
		else if (!strcasecmp(param[i], "password")){
			u_char j=strlen(param[i+1]);
			if (j < c->min_passwd_length) {
				aParse(conn, "password length is too small (real %d need at least %d)\n", j, c->min_passwd_length);
				return;
			}
			ld->password=set_string(param[i+1]);
		}
		else if (!strcasecmp(param[i], "ip")) {
			inet_aton(param[i+1], &ld->ip_from);
		}
		else if (!strcasecmp(param[i], "mac")){
			struct ether_addr *o_mac;
			o_mac=ether_aton(param[i+1]);
			if (o_mac) memcpy(&ld->mac_from, o_mac, sizeof (struct ether_addr));
		}
		else if (!strcasecmp(param[i], "strict")){
			ld->flags|=LOGIN_STRICT; i--;
		}
		else if (!strcasecmp(param[i], "nostrict")){
			ld->flags&=~LOGIN_STRICT; i--;
		}
		i+=2;
	}

	aDebug(DEBUG_LOGIN, "set/got: oid=%06X, inact:%ld, abs:%ld, pass:%s, %s, %s\n", u->id, ld->inact, ld->abs, ld->password?ld->password:"-", (ld->flags&LOGIN_ACTIVE)?"ACTIVE":"NOT ACTIVE", (ld->flags&LOGIN_STRICT)?"STRICT":"NOT STRICT");
	
	ld->flags|=LOGIN_UPDATE;
}

//////////////////////////////////////////////////////////////////////////////////////////
void sLgLoginCfg(char *param[], Connection *conn, u_char no_flag, Login_cfg *cfg){
	NetUnit *u;
	char *password=NULL;
	struct in_addr o_ip;
	struct ether_addr o_mac;
	u_char flag=no_flag?LG_LOGOUT:LG_LOGIN;

	bzero(&o_ip, sizeof (struct in_addr));
    	bzero(&o_mac, sizeof (struct ether_addr));

	u_char i=1;
	u=aParseUnit(param, &i);
	if(u==NULL) { aParse(conn, "unit not exist\n"); return; }

	if (!u->logindata) { aParse(conn, "FAIL: login service is not running for that unit\n"); return; }
	
	while (param[i]!=empty){
		if (!strcasecmp(param[i], "ip")){
			inet_aton(param[i+1], &o_ip);
		}
		else if (!strcasecmp(param[i], "mac")){
			struct ether_addr *mac;
			mac=ether_aton(param[i+1]);
			if (!memcmp(mac, "\0\0\0\0\0\1", sizeof (struct ether_addr))) flag|=LG_FORCE;
		}
		else if (!strcasecmp(param[i], "password")){
			password=param[i+1];
		}
		i+=2;
	}

#ifdef DEBUG
	char buffer[32];
	aDebug(DEBUG_LOGIN, "login/got: oid=%06X, pass:%s(%s), ip=%s, mac=%s\n", u->id, password?password:"-", u->logindata->password?u->logindata->password:"-", inet_ntop(AF_INET, &o_ip, buffer, 32), ether_ntoa(&o_mac));
#endif

	if (!password || !u->logindata->password) { 
		if (conn->permissions!=UPERM_ALL) {
			aParse(conn, "FAIL: login is not allowed\n", u->id); 
			return; 
		}
	} else 
	if (strcmp(password,u->logindata->password)) { 
		aParse(conn, "FAIL: login is not allowed\n"); 
		return; 
	}
	
	sLgLogin(conn, u, &o_ip, cfg, flag);
}
//////////////////////////////////////////////////////////////////////////////////////////
void sLgLogin(Connection *conn, NetUnit *u, struct in_addr *ip_src, Login_cfg *cfg, u_char flag) {
 	struct ether_addr zero_mac;
	struct ether_addr current_mac;
	sLoginData *ld=u->logindata;
	struct in_addr *ip;
	char buffer[32];

	bzero(&zero_mac, sizeof (struct ether_addr));
	bzero(&current_mac, sizeof (struct ether_addr));
	
	if(ip_src) ip=ip_src;
	else ip=(in_addr*)"\0\0\0\0";

	if(!(flag&LG_FORCE)) {
		// strict login processing
		if ((ld->flags&LOGIN_STRICT) && (ld->ip_from.s_addr!=ip->s_addr)) {
			aParse(conn, "FAIL: login is not allowed\n"); 
			inet_ntop(AF_INET, ip, buffer, 32);			
			LogEvent(LOGIN, u->id, 0, 0, "FAIL STRICT: login from wrong IP: %s", buffer);
			return;
		}
	
		inet_ntop(AF_INET, ip, buffer, 32);			
		aDebug(DEBUG_LOGIN, "MAC: checking %s for mac address...\n", buffer);
		sLgObtainMac(ip, &current_mac);
		memcpy(&ld->last_mac_from, &current_mac, sizeof (struct ether_addr));

		// mac address processing
		if (memcmp(&ld->mac_from, &zero_mac, sizeof (struct ether_addr))) { // unit's MAC is not zero
			if (memcmp(&ld->mac_from, &current_mac, sizeof (struct ether_addr))) { // unit's MAC != current
				aDebug(DEBUG_LOGIN, "MAC: not match database: %s\n", ether_ntoa(&ld->mac_from));
				aParse(conn, "FAIL: login is not allowed\n"); 
				LogEvent(LOGIN, u->id, 0, 0, "FAIL MAC: login from wrong mac: %s", ether_ntoa(&current_mac));
				return;
			}
		}
	} else {
		aDebug(DEBUG_LOGIN, "MAC: magic mac address is specified, check skiped\n");
	}
	
	if(flag&LG_LOGIN) { // login
		if(ld->flags&LOGIN_ACTIVE) {
			if(cfg->relogin) {
				aParse(conn, "Performing relogin\n");
				LogEvent(LOGIN, u->id, 0, 0, "RELOGIN: relogin attempt from %s\n", inet_ntop(AF_INET, ip, buffer, 32));
			} else {
				aParse(conn, "This unit already logged in\n");
				LogEvent(LOGIN, u->id, 0, 0, "FAIL LOGIN: login attempt from %s when already logged in\n",\
					inet_ntop(AF_INET, ip, buffer, 32));
				return;
			}
		}
		
		// if "set-user-ip" and LG_LOGIN, check other units currently using this IP address and logout them
		// this will use sLgLogin recursive call - hope this is OK
		if (cfg->set_user_ip){
			NetUnit *t; NetUnit_user *ut;
			for (t=Units->root; t!=NULL; t=t->next){
				if (u->type==NETUNIT_USER) {
					ut=(NetUnit_user*)t;
					if (ut->logindata) {
						if (ut->logindata->flags&LOGIN_ACTIVE && 
							!memcmp(&ut->logindata->last_ip_from, ip, sizeof (struct in_addr)) &&
							!memcmp(&ut->ip, ip, sizeof (struct in_addr)) ) {
								aDebug (DEBUG_LOGIN, "unit user %s (%06X) was logged on there, forcibly log off it\n", ut->name?ut->name:"<\?\?>", ut->id);
								sLgLogin(conn, t, NULL, cfg, LG_LOGOUT);
								break;
							}
					} 
				}
			}
		}
		
		cAccessScriptCall(DROP, u, "LOGIN OFF");

		ld->flags|=LOGIN_ACTIVE; // open the unit
		ld->opened=time(NULL);
		
		memcpy(&ld->last_ip_from, ip, sizeof (struct in_addr));
		if (cfg->set_user_ip && u->type==NETUNIT_USER) {
			u->unit2trees(REMOVE);
			NetUnit_user *uu = (NetUnit_user*)u;
			if(ip_src) memcpy(&uu->ip, ip, sizeof (struct in_addr));
			aDebug (DEBUG_LOGIN, "unit user ip address is set to %s\n", inet_ntop(AF_INET, ip, buffer, 32));
			u->unit2trees(ADD);
		}
		u->SetSysPolicy(SP_DENY_LOGIN, REMOVE);
		cAccessScriptCall(PASS, u, "LOGIN ON");
		aDebug(DEBUG_LOGIN, "login success from ip:%s, mac:%s\n", inet_ntop(AF_INET, ip, buffer, 32), ether_ntoa(&current_mac));
		aParse(conn, "OK: login success from ip:%s, mac:%s\n", inet_ntop(AF_INET, ip, buffer, 32), ether_ntoa(&current_mac));
		LogEvent(LOGIN, u->id, 0, 0, "OK login success from ip: %s mac: %s", inet_ntop(AF_INET, ip, buffer, 32), ether_ntoa(&current_mac));
	} 
	else if(flag&LG_LOGOUT) { // logout
		u->SetSysPolicy(SP_DENY_LOGIN, ADD);
		cAccessScriptCall(DROP, u, "LOGIN OFF");
		if(u->ap) u->ap->ClearLastUsed(); //clear last_used

		aDebug(DEBUG_LOGIN, "logout success from ip:%s, mac:%s\n", inet_ntop(AF_INET, ip, buffer, 32), ether_ntoa(&current_mac));
		aParse(conn, "OK: logout success from ip:%s, mac:%s\n", inet_ntop(AF_INET, ip, buffer, 32), ether_ntoa(&current_mac));
		LogEvent(LOGIN, u->id, 0, 0, "OK logout success from ip: %s mac: %s", inet_ntop(AF_INET, ip, buffer, 32), ether_ntoa(&current_mac));	
		ld->flags&=~LOGIN_ACTIVE; // close the unit
		if (cfg->set_user_ip && u->type==NETUNIT_USER) {
			u->unit2trees(REMOVE);
			NetUnit_user *uu = (NetUnit_user*)u;
			bzero(&uu->ip, sizeof (struct in_addr));
			aDebug (DEBUG_LOGIN, "unit user ip address is cleared\n");
		}
	}
	u->logindata->flags|=LOGIN_UPDATE;
}
//////////////////////////////////////////////////////////////////////////////////////////
void sLgObtainMac(struct in_addr *u, struct ether_addr *current_mac){

#if defined(FREEBSD) || defined(OPENBSD)
	
#ifdef DEBUG
	char buffer[32];
#endif
	char *buf;
	char *e;
	int mib[6];
	size_t needed;
	char *lim, *next;
	struct rt_msghdr *rtm;
	struct sockaddr_inarp *sin;
	struct sockaddr_dl *sdl;

	mib[0] = CTL_NET; mib[1] = PF_ROUTE; mib[2] = 0; mib[3] = AF_INET; mib[4] = NET_RT_FLAGS; mib[5] = RTF_LLINFO;
	if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0) return;
	if ((buf = (char*)aMalloc(needed)) == NULL) return;
	if (sysctl(mib, 6, buf, &needed, NULL, 0) < 0) return;

	lim = buf + needed;

	for (next = buf; next < lim; next += rtm->rtm_msglen) {
		rtm = (struct rt_msghdr *)next;
		sin = (struct sockaddr_inarp *)(rtm + 1);
		sdl = (struct sockaddr_dl *)((char *)sin + ROUNDUP(sin->sin_len));
		e=LLADDR(sdl);
		if (sin->sin_addr.s_addr==u->s_addr && memcmp(e, "\0\0\0\0\0\0", sizeof(struct ether_addr)))  {
			memcpy(current_mac, e, sizeof (struct ether_addr));
#ifdef DEBUG
			aDebug(DEBUG_LOGIN, "MAC for ip: %s is: %s found in ARP\n", inet_ntop(AF_INET, u, buffer, 32), ether_ntoa((struct ether_addr*)e));
#endif
			aFree(buf);
			return;
		}
	}

#ifdef DEBUG
	aDebug(DEBUG_LOGIN, "MAC for ip:%s is not found in ARP\n", inet_ntop(AF_INET, u, buffer, 32));
#endif
	aFree(buf);
#endif

#if defined(LINUX)
#ifdef DEBUG
	char buffer[32];
#endif	
	char *buf;

	char b_ip[32], b_hw[32], b_trash[32];
	struct in_addr in;
	struct ether_addr *e;
	FILE *arp;

	if ((buf = (char*)aMalloc(256)) == NULL) return;

	arp=fopen("/proc/net/arp", "rt");
	if (arp==NULL) { aDebug(DEBUG_LOGIN, "failed to open /proc/net/arp\n"); return; }

	if (NULL==fgets(buf, 255, arp)) goto FINISH_LINUX_CHECK;

	while (!feof(arp)){
		if (NULL==fgets(buf, 255, arp)) break;
		bzero(b_ip, 31); bzero(b_hw, 8);
		sscanf(buf, "%s%s%s%s", b_ip, b_trash, b_trash, b_hw);
		if (inet_aton(b_ip, &in) && (e=ether_aton(b_hw))){
			if (u->s_addr==in.s_addr && memcmp(e, "\0\0\0\0\0\0", sizeof(struct ether_addr))) {
				memcpy(current_mac, e, sizeof (struct ether_addr));
#ifdef DEBUG
				aDebug(DEBUG_LOGIN, "MAC for ip: %s is: %s found in ARP\n", inet_ntop(AF_INET, u, buffer, 32), ether_ntoa((struct ether_addr*)e));
#endif
				goto FINISH_LINUX_CHECK;
			}
		}
	}
FINISH_LINUX_CHECK:
	fclose(arp);
	aFree(buf);
#endif
	return;
}
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
