/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: s_monitor.c,v 1.63.2.7 2006/03/16 16:03:04 anton Exp $ */

#include "netams.h"

#define MONITOR_DELAY	60	//delay between load into sql

//////////////////////////////////////////////////////////////////////////////////////////
void sMonitorInit(Service *s){
	s->cfg = (Monitor_cfg*)aMalloc(sizeof(Monitor_cfg));
	Monitor_cfg *cfg=(Monitor_cfg*)s->cfg;
    	
	cfg->name=NULL;
    	cfg->total=0;
	cfg->commit=0;
	cfg->time=0;
    	cfg->to=MON_NONE;
    	cfg->storage_no=0;
	cfg->fd=NULL;

	netams_mutex_init(&cfg->lock, NULL);
}
//////////////////////////////////////////////////////////////////////////////////////////
void sMonitorProcessCfg(char *param[], Connection *conn, u_char no_flag){
	Monitor_cfg *cfg=(Monitor_cfg*)conn->service->cfg;

	if (!strcasecmp(param[0], "monitor"))
	if (!strcasecmp(param[1], "?")) { cHelp(conn, no_flag, "monitor"); return; }
	
	if (!strcasecmp(param[1], "to")) {
		Service *st;
		if (cfg->fd) {
			fclose(cfg->fd);
		}
		cfg->fd=NULL;

		if(cfg->to==MON_STORAGE) {
			st = Services->getService(SERVICE_STORAGE, cfg->storage_no);
			if(st) stSaveSql(st,cfg->name,ST_CONN_MONITOR);
		}
		
		if(cfg->name) {
			aFree(cfg->name);
			cfg->name=NULL;
		}

		if (!strcasecmp(param[2], "storage")) {
			Service *st;
			u_short no;
			no=strtol(param[3], NULL, 10);
			st = Services->getService(SERVICE_STORAGE, no);		
			if(!st) {
				aParse(conn, "no such service storage:%u\n", no);
				return;
			} 

			if(((ServiceStorage_cfg*)st->cfg)->accept[ST_CONN_MONITOR]!=1) {
				aParse(conn, "this service storage %u does not accept monitoring data\n",cfg->storage_no);
				return;
			} 

			if (no_flag) { 
				aParse(conn, "switch monitoring off\n");
				cfg->to=MON_NONE;
				return;
			}
			cfg->storage_no=no;
            		cfg->to=MON_STORAGE;
			print_to_string(&cfg->name,"monitor.%u",conn->service->instance);
			aParse(conn, "monitoring to storage %u (%p)\n", cfg->storage_no, st);
				
		} else if (!strcasecmp(param[2], "file")) {
                        if (no_flag) {
				aParse(conn, "switch monitoring off\n");
                                cfg->to=MON_NONE;
				if(cfg->name) aFree(cfg->name);
				cfg->name=NULL;
                                return;
                        }

            		cfg->to=MON_FILE;
            		u_short i=strlen(param[3]);
            		if (i) { 
                		cfg->fd=fopen(param[3], "at");
                		if (cfg->fd) {
                    			setlinebuf(cfg->fd);
					cfg->name=set_string(param[3]);
        				aParse(conn, "monitoring to file %s\n", cfg->name);
                		} else { 
                    			aParse(conn, "unable to open monitor output file %s\n", cfg->name);
                    			cfg->to=MON_NONE;
                    			cfg->fd=NULL;
                    		}
                	}
            		else 
				aParse(conn, "monitor output file name unknown\n");
		} else 
			aParse(conn, "unknown monitor output method\n");
	} else if (!strcasecmp(param[1], "unit")) {
        	NetUnit *u;
		u_char i=2;
		
		u = aParseUnit(param, &i);
        	if (u) { 
            		if (!no_flag) { 
                		if (Units->setMon(u->id, cfg)) 
					aParse(conn, "monitoring unit %s (%06X)\n", u->name?u->name:"<\?\?>", u->id);
                		else 
					aParse(conn, "set monitoring unit %06X failed!\n", u->id);
            		} else {
                		if (Units->setMon(u->id, NULL)) 
					aParse(conn, "not monitoring unit %s (%06X)\n", u->name?u->name:"<\?\?>", u->id);
                		else 
					aParse(conn, "unset monitoring unit %06X failed!\n", u->id);
                	}
        	} else {
            		if (no_flag) {
                		Units->setMon(0, cfg);
                		aParse(conn, "disable monitoring for all units\n");
            		} else aParse(conn, "cannot monitor this: unit unknown\n");
       		}
    	}

	else aParse(conn, "unknown monitor command\n");

   }	

//////////////////////////////////////////////////////////////////////////////////////////
void cShowMonitor(Connection *conn){
	Service *s=NULL;
        
	while((s=Services->getServiceNextByType(SERVICE_MONITOR,s))) {
        	Monitor_cfg *cfg=(Monitor_cfg*)s->cfg;

		fprintf(conn->stream_w, "service monitor %u\n",s->instance);
    		switch (cfg->to){
        		case MON_NONE:  
				fprintf(conn->stream_w, "Monitoring is off\n");
                    	break;
        	case MON_FILE:  
				fprintf(conn->stream_w, "Monitoring to file: %s\n", cfg->name);
                    		break;
        	case MON_STORAGE:   
				fprintf(conn->stream_w, "Monitoring to storage: %u\n", cfg->storage_no);
                        	break;
        	}
    		if (cfg->to!=MON_NONE) { 
        		fprintf(conn->stream_w, "Units: ");
        		Units->showMon(conn->stream_w, (void*)cfg, 1);
        		fprintf(conn->stream_w, "\n");
        	}
		fprintf(conn->stream_w, "Packets monitored: %lu\n",cfg->total);
	}
    }

//////////////////////////////////////////////////////////////////////////////////////////
void sMonitorListCfg(Service *s, FILE *out){
        Monitor_cfg *cfg=(Monitor_cfg*)s->cfg;
    	
	switch (cfg->to){
        	case MON_NONE:  return;
        	case MON_FILE:  {
                    	fprintf(out, "monitor to file %s\n", cfg->name);
                    	break;
                    	}
        	case MON_STORAGE:   {
                    	fprintf(out, "monitor to storage %u\n", cfg->storage_no);
                    	break;
                    	}
        	}
   		if (cfg->to!=MON_NONE) Units->showMon(out, (void*)cfg, 0);
    	
	fprintf(out, "\n");

}

//////////////////////////////////////////////////////////////////////////////////////////
void aMonitor(NetUnit *u, Flow *flow){
	Monitor_cfg *cfg=(Monitor_cfg*)u->monitor;

	if (!cfg) {
		Service *sm=Services->getServiceNextByType(SERVICE_MONITOR, NULL);
		if (sm) cfg=(Monitor_cfg*)sm->cfg;
		else return; // no place to monitor to
	}
	
	struct flow_info_value		*flow_info	= (struct flow_info_value*)flow->get(ATTR_FLOW_INFO);
        struct ipv4_info_value		*ipv4_info	= (struct ipv4_info_value*)flow->get(ATTR_IPV4_INFO);
	struct ifindex_info_value	*ifindex_info	= (struct ifindex_info_value*)flow->get(ATTR_IFINDEX_INFO);
	struct as_info_value		*as_info	= (struct as_info_value*)flow->get(ATTR_AS_INFO);
	struct tcp_info_value		*tcp_info	= (struct tcp_info_value*)flow->get(ATTR_TCP_INFO);
	struct layer7_info_value		*layer7_info	= (struct layer7_info_value*)flow->get(ATTR_LAYER7_INFO);
	
	unsigned long t	= flow_info->flow_last;
	unsigned flowt	=(unsigned)(flow_info->flow_last - flow_info->flow_first);
	
#ifdef DEBUG
	if(aDebug(DEBUG_MONITOR,"sending flow %p \n", flow)) {
		flow->Debug(DEBUG_MONITOR);
	}
#endif

	netams_mutex_lock(&cfg->lock);

	if (cfg->to==MON_FILE) {
		char t_T[30];
		timeU2T( t, (char*)&t_T);

		fprintf(cfg->fd, "%s %u %s %06X ", (char*)&t_T, flowt, u->name?u->name:"<\?\?>", u->id);
		
		if(ipv4_info) {
			char buf1[32], buf2[32];
			inet_ntop(AF_INET, &(ipv4_info->ip_src), buf1, 32);
			inet_ntop(AF_INET, &(ipv4_info->ip_dst), buf2, 32);
     			
			fprintf(cfg->fd, "%02u ", ipv4_info->ip_p);
		
			if (tcp_info) {
    	    			fprintf(cfg->fd, "s:%s:%u d:%s:%u ",
					buf1, ntohs(tcp_info->src_port),
					buf2, ntohs(tcp_info->dst_port));
        		} else 
    	    			fprintf(cfg->fd, "s:%s d:%s ", buf1, buf2);
		}

		if(ifindex_info)
			fprintf(cfg->fd, "if:%u->%u ", ntohs(ifindex_info->if_in), ntohs(ifindex_info->if_out));
		if(as_info)
			fprintf(cfg->fd, "as:%u->%u ", ntohs(as_info->as_src), ntohs(as_info->as_dst));
		if (layer7_info)
			if (layer7_info->value)
				fprintf(cfg->fd, "url:%s ", layer7_info->value);

		fprintf(cfg->fd, " %lu %llu\n", flow_info->packets, flow_info->octets);
	
	} else if (cfg->to==MON_STORAGE) {
		if(!ipv4_info) {
			aLog(D_WARN, "Trying to monitor non IPv4 packet, not enough code bits\n");
			goto EXIT;
		}

		if(!cfg->fd) {
			cfg->fd=fopen(cfg->name,"a");
			if(!cfg->fd) {
				aLog(D_ERR, "Can't open temporary file %s : %s\n",cfg->name,strerror(errno));
				goto EXIT;
			}
			setlinebuf(cfg->fd);
		}
		
		fprintf(cfg->fd, "%lu,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%u,%lu,%llu,%s\n",
			t,
			flowt,
			u->id,
			ipv4_info->ip_p,
			(unsigned)ntohl(ipv4_info->ip_src.s_addr),
			tcp_info?ntohs(tcp_info->src_port):0,
			(unsigned)ntohl(ipv4_info->ip_dst.s_addr),
			tcp_info?ntohs(tcp_info->dst_port):0,
			ifindex_info?ntohs(ifindex_info->if_in):0,
			ifindex_info?ntohs(ifindex_info->if_out):0,
			as_info?ntohs(as_info->as_src):0,
			as_info?ntohs(as_info->as_dst):0,
			flow_info->packets,
			flow_info->octets,
			layer7_info?(layer7_info->value?layer7_info->value:""):"");
		
		cfg->commit++;
		if( cfg->time < t)  {
			cfg->time=t+MONITOR_DELAY;
                	Service *st = Services->getService(SERVICE_STORAGE, cfg->storage_no);
                	if(st) {
				fclose(cfg->fd);
				cfg->fd=NULL;
				stSaveSql(st,cfg->name,ST_CONN_MONITOR);
			}
                	else goto EXIT;
			cfg->commit=0;
        	} 
        }
	cfg->total++;

	netams_mutex_unlock(&cfg->lock);

EXIT: ;
}
//////////////////////////////////////////////////////////////////////////////////////////
void sMonitorCancel(void *v){
        Service *s = (Service*)v;
        Monitor_cfg *cfg=(Monitor_cfg*)s->cfg;
	
	aLog(D_INFO, "cancelling service monitor:%u\n", s->instance);
	
	Units->setMon(0, cfg);
	
	if(cfg->fd) fclose(cfg->fd);
	if(cfg->to==MON_STORAGE) {
		Service *st = Services->getService(SERVICE_STORAGE, cfg->storage_no);
                if(st) {
			stSaveSql(st, cfg->name, ST_CONN_MONITOR);
        		stCloseSql(st,ST_CONN_MONITOR);
		}
	}
	if(cfg->name) aFree(cfg->name);

	netams_mutex_destroy(&cfg->lock);
}

//////////////////////////////////////////////////////////////////////////////////////////
