/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: s_processor.c,v 1.158.2.9 2006/05/22 10:15:20 jura Exp $ */

#include "netams.h"

void *sProcessor(void *ss);
void sProcessorCancel(void *);
void sProcessorListCfgUnit(NetUnit *d, FILE *f, u_char flags);
void sPConstructStoreMessages(oid netunit, policy_data *pd);
void sPMessagesGenerator(ServiceProcessor_cfg *cfg);
void sPMessagesMultiplexer(ServiceProcessor_cfg *cfg);

int cAutoAssignIP(Connection *conn, char *param[], u_char no_flag);
int cAutoUnits(Connection *conn, char *param[], u_char no_flag);
int cDefault(Connection *conn, char *param[], u_char no_flag);
int cRestrict(Connection *conn, char *param[], u_char no_flag);

NetUnitsList *Units;
PolicyList *PolicyL;

FIFO *Mux_in;
Service *Processor;
//////////////////////////////////////////////////////////////////////////
ServiceProcessor_cfg ProcessorCfg;
//////////////////////////////////////////////////////////////////////////
void sProcessorInit(Service *s){
	ServiceProcessor_cfg *cfg=&ProcessorCfg;
	cfg->def=NULL;
	cfg->delay=PROCESSOR_DELAY; // in seconds
	cfg->lifetime=PROCESSOR_LIFETIME;
	cfg->restrict_all=DROP;
	cfg->restrict_local=PASS;
	cfg->access_script=NULL;
	cfg->auto_assign=NULL;
	cfg->auto_units=NULL;
	
	Mux_in = new FIFO(MAX_UNITS*5);

	cfg->mac_control_units_checked=0;
	cfg->mac_control_units_violated=0;
	//this pointers might be defined from others places, remember it
//	cfg->fifo=NULL;
//	cfg->st_root=NULL;

	Processor=s;
	s->cfg=(void*)cfg;
	pthread_create(&(s->t_id), NULL, &sProcessor, s);
}
//////////////////////////////////////////////////////////////////////////
void sProcessorProcessCfg(char *param[], Connection *conn, u_char no_flag){
	ServiceProcessor_cfg *cfg=(ServiceProcessor_cfg*)conn->service->cfg;

	if (!strcasecmp(param[0], "unit")) cUnit(conn, param, no_flag);
	else if (!strcasecmp(param[0], "policy")) cPolicy(conn, param, no_flag);
	else if (!strcasecmp(param[0], "default")) cDefault(conn, param, no_flag);
	else if (!strcasecmp(param[0], "restrict")) cRestrict(conn, param, no_flag);
	else if (!strcasecmp(param[0], "lookup-delay")) {
		unsigned delay=strtol(param[1], NULL, 10);
		if (delay>=1 && delay<24*60*60) {
			aParse(conn, "lookup delay is set to %u seconds\n", delay);
			cfg->delay=delay;
		} 
		else aParse(conn, "lookup delay value invalid\n");
	}		
	else if (!strcasecmp(param[0], "flow-lifetime")) {
		unsigned lifetime=strtol(param[1], NULL, 10);
		if (lifetime>=1 && lifetime<24*60*60) {
			aParse(conn, "flow lifetime is set to %u seconds\n", lifetime);
			cfg->lifetime=lifetime;
		} 
		else aParse(conn, "flow lifetime value invalid\n");
	}		
	else if (!strcasecmp(param[0], "access-script")) {
		if(cfg->access_script) aFree(cfg->access_script);
		cfg->access_script=set_string(param[1]);
		aParse(conn, "access control script name is set to '%s'\n", param[1]);
	}		
	else if (!strcasecmp(param[0], "auto-assign")) cAutoAssignIP(conn, param, no_flag);
	else if (!strcasecmp(param[0], "auto-units")) cAutoUnits(conn, param, no_flag);
	else aParse(conn, "unknown processor command: %s\n", param[0]);
}
//////////////////////////////////////////////////////////////////////////
void *sProcessor(void *ss){
	Service *s=(Service*)ss;
	ServiceProcessor_cfg *cfg=(ServiceProcessor_cfg*)s->cfg;
	unsigned long long loop_iteration=0;
	Service *dependent=NULL;

	aLog(D_INFO, "service processor thread started\n");
	pthread_cleanup_push(sProcessorCancel, s);
	SET_CANCEL();
	Processor->t_id=pthread_self();
	
	// now we should sleep before starting actual processor
	if(!(s->flags&SERVICE_RUN)) Processor->Sleep();

	Service* st = aStorageGetAccepted(ST_CONN_SUMMARY);
	if (!st) aLog(D_WARN, "no storages with SUMMARY data are available\n");
	else aLog(D_INFO, "will use storage %d for SUMMARY data\n", st->instance);
	
	Processor->Sleep(1+cfg->delay/10);
	// loop for incoming data forever
	while(1) {
		//generate message for units
		sPMessagesGenerator(cfg);	
		
		//deliver generated messages to storages
		sPMessagesMultiplexer(cfg);

		// detect data-dependent services and wake up them
		dependent=Services->getServiceNextByType(SERVICE_LOGIN, NULL); if (dependent) dependent->Wakeup();
		dependent=Services->getServiceNextByType(SERVICE_QUOTA, NULL); if (dependent) dependent->Wakeup();
		dependent=Services->getServiceNextByType(SERVICE_BILLING, NULL); if (dependent) dependent->Wakeup();

		Processor->Sleep(cfg->delay);
		loop_iteration++;
	} // infinite while
	// we will never reach this point
	pthread_cleanup_pop(0);
	return NULL;
}
//////////////////////////////////////////////////////////////////////////
void sProcessorCancel(void *v){
	Service *s=(Service*)v;
	ServiceProcessor_cfg *cfg=(ServiceProcessor_cfg*)s->cfg;
	
	cfg->lifetime=0; // this will flush all data when we will generate messages
	sPMessagesGenerator(cfg);
	aLog(D_INFO, "cancelling service processor, flushing %u messages\n", Mux_in->num_items);
	//deliver generated messages to storages
	sPMessagesMultiplexer(cfg);
	
	//destroy Multiplexer fifo
	delete Mux_in;
	
	Processor=NULL;
	
	//clar AutoAssign config
	AutoAssignEntry *ase;
	while(cfg->auto_assign) {
		ase=cfg->auto_assign;
		cfg->auto_assign=cfg->auto_assign->next;
		aFree(ase);
	}

	//clear AutoUnits config
	AutoUnitsEntry *aue;
	while(cfg->auto_units) {
		aue=cfg->auto_units;
		cfg->auto_units=cfg->auto_units->next;
		aFree(aue->put_to_group);
		aFree(aue->prefix);
		aFree(aue);
	}

	if(cfg->access_script) aFree(cfg->access_script);
	if(cfg->def) delete cfg->def;
	
	//clear it!
	s->cfg=NULL;
	bzero(&ProcessorCfg, sizeof(ServiceProcessor_cfg));
}
//////////////////////////////////////////////////////////////////////////
void sProcessorListCfg(Service *s, FILE *f, u_char flags){
	ServiceProcessor_cfg *cfg=(ServiceProcessor_cfg*)s->cfg;
	
	if(cfg->delay != PROCESSOR_DELAY) fprintf(f, "lookup-delay %d\n", cfg->delay);
	if(cfg->lifetime != PROCESSOR_LIFETIME) fprintf(f, "flow-lifetime %d\n", cfg->lifetime);
	
	char buf[256];

	// first, policy rules
	netams_rwlock_rdlock(&PolicyL->rwlock);
	for (Policy *d=PolicyL->root; d!=NULL; d=d->next) {
		fprintf(f, "policy ");

		fprintf(f, "oid %06X ", d->id);
		if (d->name) SHOW_STR(f, "name", d->name);

		if (d->target.target_type) {
			fprintf(f, "target %s", d->getTarget(buf));
		}
		if (d->bwi) {
			fprintf(f, "bw %s", getBW(d->bwi, buf));
		}
		fprintf(f, "\n");
	}
	netams_rwlock_unlock(&PolicyL->rwlock);

	// default parameters
	if (cfg->def && cfg->def->ap) { fprintf(f, "default acct-policy "); cfg->def->ap->ListForCfg(f, flags); }
	if (cfg->def && cfg->def->fp) { fprintf(f, "default fw-policy "); cfg->def->fp->ListForCfg(f, flags); } 

	// restrict parameters
	fprintf(f, "restrict all %s local %s\n", cfg->restrict_all?"drop":"pass", cfg->restrict_local?"drop":"pass");

	// auto-assign ip addresses
	for(AutoAssignEntry *e=cfg->auto_assign; e!=NULL; e=e->next ) {
		fprintf(f, "auto-assign %s", inet_ntop(AF_INET, &(e->start), buf, 32));
		fprintf(f, " %s\n", inet_ntop(AF_INET, &(e->stop), buf, 32));
	}

	// auto-units 
	for(AutoUnitsEntry *e=cfg->auto_units; e!=NULL; e=e->next ) {
		fprintf(f, "auto-units %u ", e->id);
		if (e->naming==AU_NAMING_BY_DNS) 
			fprintf(f, "type %s naming by-dns ", e->type==AU_TYPE_HOST?"host":"user");
		else 
			fprintf(f, "type %s naming prefix%u %s ", e->type==AU_TYPE_HOST?"host":"user", e->naming, e->prefix);
		if (e->put_to_group) if (Units->getUnit(e->put_to_group)) fprintf(f, "group %s", e->put_to_group);
		fprintf(f, "\n");
	}

	// next, print NetUnits
	netams_rwlock_rdlock(&Units->rwlock);
	NetUnit *d;
	for (d=Units->root; d!=NULL && d->type==NETUNIT_GROUP; d=d->next) {
		fprintf(f, "unit group oid %06X ", d->id);
		if (d->name) SHOW_STR(f, "name", d->name);
		if (d->description) SHOW_STR(f, "description", d->name);
		sProcessorListCfgUnit(d, f, flags);
		fprintf(f, "\n");
	} //for
	
	for (; d!=NULL; d=d->next) {
		fprintf(f, "unit %s oid %06X ", netunit_type_name[d->type], d->id);
		if (d->name) SHOW_STR(f, "name", d->name);

		switch (d->type) {
			case NETUNIT_HOST: {
				NetUnit_host *h = (NetUnit_host*)d;
				fprintf(f, "ip %s ", inet_ntop(AF_INET, &(h->ip), buf, 32));
				if (h->mac) fprintf(f, "mac %s ", ether_ntoa(h->mac));
				break;
			}
			case NETUNIT_CLUSTER: {
				NetUnit_net *t;
				NetUnit_cluster *h = (NetUnit_cluster*)d;
				for (t=h->root; t!=NULL; t=t->next)
					fprintf(f, "ip %s/%u ", inet_ntop(AF_INET, &(t->ip), buf, 32), MASK2MLEN(t->mask));
				break;
			}
			case NETUNIT_NET: {
				NetUnit_net *h = (NetUnit_net*)d;
				fprintf(f, "ip %s ", inet_ntop(AF_INET, &(h->ip), buf, 32));	
				fprintf(f, "mask %s ", inet_ntop(AF_INET, &(h->mask), buf, 32));
				if (h->auto_units_id) fprintf(f, "auto-units %u ", h->auto_units_id);
				break;
			}
			case NETUNIT_USER: {
				NetUnit_user *u = (NetUnit_user*)d;
				if(u->ip.s_addr!=0) fprintf(f, "ip %s ", inet_ntop(AF_INET, &(u->ip), buf, 32));
				if (u->mac) fprintf(f, "mac %s ", ether_ntoa(u->mac));
				break;
			}
			default: 
				continue;
		} //switch
		if(d->description) if (strchr(d->description, ' ')) 
				fprintf(f, "description \"%s\" ", d->description);
			else fprintf(f, "description %s ", d->description);

		sProcessorListCfgUnit(d, f, flags);
		fprintf(f, "\n");
	} //for
	netams_rwlock_unlock(&Units->rwlock);
	
	if (cfg->access_script) fprintf(f, "access-script \"%s\"\n", cfg->access_script);

	fprintf(f, "\n");
}
//////////////////////////////////////////////////////////////////////////
void sProcessorListCfgUnit(NetUnit *d, FILE *f, u_char flags){
	char buf[256];
	policy_data *cpd;

	if (d->email) fprintf(f, "email %s ", d->email);
	if (d->password) fprintf(f, "password %s ", (flags&CFG_NO_PASSWORDS)?"***":d->password);

	//bw info
	if (d->bwi) fprintf(f, "bw %s ", getBW(d->bwi, buf));

	DSList *dsl=d->getDSList();
	if (dsl){
		fprintf(f, "ds-list ");
		for (; dsl!=NULL; dsl=dsl->next) {
			if (dsl!=d->dsl_root) fprintf(f, ",");
			fprintf(f, "%u", dsl->id);
		}
		fprintf(f, " ");
	}

	if (d->parent && d->parent->name) {
		fprintf(f, "parent ");
		SHOW_OIDS(f, flags, d->parent->name, d->parent->id);
	}
	if (d->sys_policy!=SP_NONE) { 
		//in config print only few sys_policies
		PrintSysPolicy(f, d->sys_policy&SP_DENY,d->sys_policy_perm);    
	}
	if (d->flags&NETUNIT_NLP) fprintf(f, "no-local-pass ");
	if (d->flags&NETUNIT_BACKUP_FW) fprintf(f, "backup-fw ");

	if (d->ap
#ifdef HAVE_BILLING
	&& !d->account
#endif
	) {
		u_char header_printed=0, already_default; 
		policy_data *cpx;

		for (cpd=d->ap->root; cpd!=NULL; cpd=cpd->next) {
			already_default=0;
			if (ProcessorCfg.def && ProcessorCfg.def->ap) {
				for (cpx=ProcessorCfg.def->ap->root; cpx!=NULL; cpx=cpx->next) 
					if (cpx->policy->id==cpd->policy->id) {
						already_default=1;
						break;
					}
			}
			if (!already_default) {
				if (!header_printed) { fprintf(f, "acct-policy "); header_printed=1; }
				fprintf(f, "%s%s",
					(cpd->policy_flags&POLICY_FLAG_INV)?"!":"", 
					(cpd->policy_flags&POLICY_FLAG_BRK)?"%":"");
				SHOW_OIDS(f, flags, cpd->policy->name, cpd->policy->id);
			}
		}
	}
	if (d->fp) { 
		u_char header_printed=0, already_default; 
		policy_data *cpx;

		for (cpd=d->fp->root; cpd!=NULL; cpd=cpd->next) {
			already_default=0;
			if (ProcessorCfg.def && ProcessorCfg.def->fp) {
				for (cpx=ProcessorCfg.def->fp->root; cpx!=NULL; cpx=cpx->next) 
					if (cpx->policy->id==cpd->policy->id) {
						already_default=1;
						break;
					}
			}
			if (!already_default) {
				if (!header_printed) { fprintf(f, "fw-policy "); header_printed=1; }
				fprintf(f, "%s%s", 
					(cpd->policy_flags&POLICY_FLAG_INV)?"!":"",
					(cpd->policy_flags&POLICY_FLAG_BRK)?"%":"");
				SHOW_OIDS(f, flags, cpd->policy->name, cpd->policy->id);
			}
		}
	}
}
//////////////////////////////////////////////////////////////////////////
void cShowProcessor(Connection *conn){
	ServiceProcessor_cfg *cfg=&ProcessorCfg;
	u_short hld_size=sizeof(MsgHolder);

	MsgMgr->Usage(conn);

	fprintf(conn->stream_w, "\nINPUT  Multiplexer\n");
	fprintf(conn->stream_w, "\t current: %u\tmax: %u\ttotal: %lu\t(%ub)\n", Mux_in->num_items, Mux_in->max_items, Mux_in->total_items,Mux_in->num_holders*hld_size); 
	fprintf(conn->stream_w, "OUTPUT Multiplexers\n");

	storage_list *item, *root=NULL;
	aStorageGetAcceptedAll(ST_CONN_SUMMARY, &root);
	aStorageGetAcceptedAll(ST_CONN_RAW, &root);

	for (item=root; item!=NULL; item=item->next) {
		fprintf(conn->stream_w, "   Storage %d type ", item->s->instance);
		switch (item->cfg->type) {
			case HASH: fprintf(conn->stream_w, "HASH"); break;
			case MY_SQL: fprintf(conn->stream_w, "MySQL"); break;
			case POSTGRES: fprintf(conn->stream_w, "Postgres"); break;
			case ORACLE: fprintf(conn->stream_w, "Oracle"); break;
			case LIBRADIUS: fprintf(conn->stream_w, "Radius"); break;
			default: fprintf(conn->stream_w, "UNKNOWN"); break;
		}
		fprintf(conn->stream_w, "\n");
		fprintf(conn->stream_w, "\t in current: %u\tmax: %u\ttotal: %lu\t(%ub)\n", item->cfg->in->num_items, item->cfg->in->max_items, item->cfg->in->total_items, item->cfg->in->num_holders*hld_size); 
		fprintf(conn->stream_w, "\tout current: %u\tmax: %u\ttotal: %lu\t(%ub)\n", item->cfg->out->num_items,	item->cfg->out->max_items, item->cfg->out->total_items, item->cfg->out->num_holders*hld_size); 
	}
	for (item=root; root!=NULL; item=root) {
		root=item->next;
		aFree(item);
	}
	
	if(cfg->fifo) {
		fprintf(conn->stream_w, "   Billing\n");
		fprintf(conn->stream_w, "\t current: %u\tmax: %u\ttotal: %lu\t(%ub)\n", cfg->fifo->num_items, cfg->fifo->max_items, cfg->fifo->total_items, cfg->fifo->num_holders*hld_size);
	}	
	
	if(cfg->def) {
		if(cfg->def->ap) {
			fprintf(conn->stream_w, "Default acct policy: ");
			cfg->def->ap->List(conn);
		}
		if(cfg->def->fp) {
			fprintf(conn->stream_w, "Default   fw policy: ");
			cfg->def->fp->List(conn);
		}
	}
	if (cfg->mac_control_units_checked) {
		fprintf(conn->stream_w, "MAC address control checked %llu, violated %llu\n", cfg->mac_control_units_checked, cfg->mac_control_units_violated);
	}
}
//////////////////////////////////////////////////////////////////////////
void sPConstructStoreOidMessage(NetUnit *u) {
	Message_Store *msgs;
	msgs = (Message_Store*)MsgMgr->New(MSG_STORE);
	msgs->ap=0;
	msgs->netunit=u->id;
	msgs->prefix='O';
	Mux_in->Push((Message*)msgs);
	aDebug(DEBUG_PROC_MUX, "P<-P O unit:%06X store oid\n", msgs->netunit);

	return;
}
//////////////////////////////////////////////////////////////////////////
void sPConstructStoreMessages(oid netunit, policy_data *pd){
	Message_Store *msg;
	
	for(u_char i=0;i<5;i++) {
		msg=(Message_Store*)MsgMgr->New(MSG_STORE);
		msg->ts=pd->to;
		msg->ap=pd->policy->id;
		msg->netunit=netunit;
		msg->prefix=pstat_prefix[i];
		switch(pstat_prefix[i]){
			case 'F':
				memcpy(msg->data, (struct pstat *)&pd->flow, sizeof (struct pstat));
				break;
			case 'M': 
				memcpy(msg->data, (struct pstat *)&pd->m, sizeof (struct pstat)); 
				break;
			case 'W':
				memcpy(msg->data, (struct pstat *)&pd->w, sizeof (struct pstat)); 
				break;
			case 'D': 
				memcpy(msg->data, (struct pstat *)&pd->d, sizeof (struct pstat)); 
				break;
			case 'H': 
				memcpy(msg->data, (struct pstat *)&pd->h, sizeof (struct pstat)); 
				break;
		}
		aDebug(DEBUG_PROC_MUX, "DS->P %c in unit:%06X acct:%06X from:%lu to:%lu in:%llu out:%llu\n",msg->prefix, msg->netunit, msg->ap, msg->data->from, msg->ts, msg->data->in, msg->data->out);
		Mux_in->Push((Message*)msg);
	}
}
//////////////////////////////////////////////////////////////////////////
void sPConstructReadMessages(NetUnit *u, policy_data *pdata) {
	Message_Read *msg;

	msg=(Message_Read*)MsgMgr->New(MSG_READ);
	msg->pdata=pdata;
	msg->ap=pdata->policy->id;
	msg->netunit=u->id;
	Mux_in->Push((Message*)msg);
	
	pdata->flags|=POLICY_DATA_LOCKED;

	aDebug(DEBUG_PROC_MUX, "P<-P H,D,W,M unit:%06X acct:%06X from:%lu\n", msg->netunit, msg->ap, msg->pdata->h.from);

	return;
}
//////////////////////////////////////////////////////////////////////////
int cAutoAssignIP(Connection *conn, char *param[], u_char no_flag) {
	in_addr start;
	in_addr stop;
	AutoAssignEntry *e,*p=NULL;
	
	if (param[1]) inet_aton(param[1], &start);
	if (param[2]) inet_aton(param[2], &stop);
	
	for(e=ProcessorCfg.auto_assign; e!=NULL; e=e->next) {
		if(e->start.s_addr==start.s_addr && e->stop.s_addr==stop.s_addr) break;
		p=e;
	}

	if(no_flag) {
		if(!e) return PARSE_OK; // nothing to remove
		if(ProcessorCfg.auto_assign==e) ProcessorCfg.auto_assign=e->next;
		else p->next=e->next;
		aFree(e);
		aParse(conn, "auto-assign %s - %s removed\n", param[1], param[2]);
	} else {
		if(e) return PARSE_OK; // already exist
		e=(AutoAssignEntry*)aMalloc(sizeof(AutoAssignEntry));
		e->start.s_addr=start.s_addr;
		e->stop.s_addr=stop.s_addr;
		if(ProcessorCfg.auto_assign==NULL) ProcessorCfg.auto_assign=e;
		else p->next=e;
		e->next=NULL;

		char buf1[32], buf2[32];
		aParse(conn, "auto-assign from %s to %s\n", inet_ntop(AF_INET, &start, buf1, 32), inet_ntop(AF_INET, &stop, buf2, 32));
        }
	
	return PARSE_OK;
}
//////////////////////////////////////////////////////////////////////////
int cAutoUnits(Connection *conn, char *param[], u_char no_flag) {
	AutoUnitsEntry *e,*p=NULL;

	u_char id=strtol(param[1], NULL, 10);
	if (!id) return 0;

	for(e=ProcessorCfg.auto_units; e!=NULL; e=e->next) {
		if(e->id==id) break;
		p=e;
	}

	if (no_flag) {
		if(!e) return PARSE_OK; // nothing to remove
		if(ProcessorCfg.auto_units==e) ProcessorCfg.auto_units=e->next;
		else p->next=e->next;
		aFree(e->put_to_group);
		aFree(e->prefix);
		aFree(e);
		aParse(conn, "auto-units %d removed\n", id);
	} else {
		if(!e)
			e = (AutoUnitsEntry*)aMalloc(sizeof(AutoUnitsEntry));
 
		if (!strcasecmp("type", param[2])) {
			if (!strcasecmp("host", param[3])) e->type=AU_TYPE_HOST; 
			else if (!strcasecmp("user", param[3])) e->type=AU_TYPE_USER; 
		}

		int k=6;
		if (!strcasecmp("naming", param[4])) {
			if (!strcasecmp("by-dns", param[5])) e->naming=AU_NAMING_BY_DNS; 
			else {
				if (!strcasecmp("prefix1", param[5])) e->naming=AU_NAMING_PREFIX1; 
				else if (!strcasecmp("prefix2", param[5])) e->naming=AU_NAMING_PREFIX2; 
				if (param[6] != empty) { e->prefix=set_string(param[6]); k++; }
			}
		}

		// we cannot deal with group ID or check group name since units are not defined yet
		if (!strcasecmp("group", param[k])) if (param[k+1]!=empty) {
			e->put_to_group=set_string(param[k+1]);
		}
	
		if(e->id) 
			aParse(conn, "auto-units %u modified, type %u, naming %u, gr=%s\n", id, e->type, e->naming, e->put_to_group?e->put_to_group:"(no)");
		else {
			e->id=id;
			if(ProcessorCfg.auto_units==NULL) ProcessorCfg.auto_units=e;
			else p->next=e;
			e->next=NULL;
			aParse(conn, "auto-units %u added, type %u, naming %u, gr=%s\n", id, e->type, e->naming, e->put_to_group?e->put_to_group:"(no)");
        	}
	}

	return PARSE_OK;
}

void CreateAutoUnit(oid parent_id, in_addr addr) {
	AutoUnitsEntry *e;
	NetUnit *net=Units->getUnitById(parent_id);

	if(!net) return;
	
	//we need this to protect parent being removed in operation
	netams_rwlock_rdlock(&Units->rwlock);
	
	u_char id=((NetUnit_net*)net)->auto_units_id;
	
        for(e=ProcessorCfg.auto_units; e!=NULL; e=e->next) {
                if(e->id==id) break;
        }
	
	if(!e) {
		netams_rwlock_unlock(&Units->rwlock);
		return;
	}

	NetUnit *u=NULL;
	au_type_enum type = e->type;
        au_naming_enum naming = e->naming;
        char *prefix = e->prefix;

	if (type == AU_TYPE_HOST) {
		u=new NetUnit_host();
		((NetUnit_host*)u)->ip.s_addr=addr.s_addr;
                                
	}
	else if (type == AU_TYPE_USER) {
		u=new NetUnit_user();
		((NetUnit_user*)u)->ip.s_addr=addr.s_addr;
                                
	}
	
	//set acct and fw policy acording of parent net 
	if(net->ap) net->ap->SetForUnit(POLICY_ACCT, u);
	if(net->fp) net->fp->SetForUnit(POLICY_FW, u);

	netams_rwlock_unlock(&Units->rwlock);
	
	u->id=newOid(0);
	char a[32], *b;
	char buf[32];
	
	inet_ntop(AF_INET, &addr, a, 32);

	aLog(D_WARN, "auto-creating unit %06X, ip=\"%s\"\n", u->id, a);

	if (e->put_to_group) {
		NetUnit *parent = Units->getUnit(e->put_to_group);
		if (parent && parent->type==NETUNIT_GROUP) 
			Units->Unit2Group(u, (NetUnit_group*)parent, ADD);
	}
	                             
	if (naming == AU_NAMING_BY_DNS) {
		struct hostent *hp;
		hp=gethostbyaddr((const char *)&addr, sizeof addr, AF_INET);
		if (hp) {
			u->name=set_string(hp->h_name);
		} 
		else { // DNS failed!
			print_to_string( &u->name, "%s", a);
		}
	}
	else if (naming == AU_NAMING_PREFIX1) {
		strncpy(buf, strrchr(a, '.')+1, 3);
		print_to_string( &u->name, "%s%s", prefix, buf);
	}
	else if (naming == AU_NAMING_PREFIX2) {
		b=strchr(a, '.');
		strncpy(buf, strchr(b+1, '.')+1, 7);
		print_to_string( &u->name, "%s%s", prefix, buf);
	}
	
	//copy ds list
	for(DSList *dsl = net->dsl_root;dsl!=NULL;dsl=dsl->next)
		u->setDSList(dsl->id);


	aLog(D_WARN, "auto-creating unit %s (%06X) for net %s (%06X)\n", u->name, u->id, net->name, net->id);
	if (u->parent) aLog(D_WARN, "\tthis unit %s (%06X) was put to group %s (%06X)\n", u->name, u->id, u->parent->name?u->parent->name:"<\?\?>", u->parent->id);

	u->unit2trees(ADD);
	Units->Insert(u);
}
//////////////////////////////////////////////////////////////////////////
int cDefault(Connection *conn, char *param[], u_char no_flag){

	u_char i=1;

	if (!strcmp(param[1], "acct-policy")) {
		if(!ProcessorCfg.def) ProcessorCfg.def = new NetUnit(NETUNIT_UNKNOWN);
		PolicyAdd(ProcessorCfg.def, &i, POLICY_ACCT, conn, param, no_flag);
	}
	else if (!strcmp(param[1], "fw-policy")) {
		if(!ProcessorCfg.def) ProcessorCfg.def = new NetUnit(NETUNIT_UNKNOWN);
		PolicyAdd(ProcessorCfg.def, &i, POLICY_FW, conn, param, no_flag);
	}
	else { aParse(conn, "default action name unspecified\n", param[1]); return PARSE_OK; }
	
	aParse(conn, "WARNING: New default policies begin to work AFTER reload!!!\n");

	return PARSE_OK;
}
//////////////////////////////////////////////////////////////////////////
int cRestrict(Connection *conn, char *param[], u_char no_flag){
	ServiceProcessor_cfg *cfg=(ServiceProcessor_cfg*)conn->service->cfg;
	u_char i=1;

	while (param[i]!=empty) {
		if (!strcmp(param[i], "all")) {
			if (!strcmp(param[i+1], "drop")){
				cfg->restrict_all=DROP;
				aParse(conn, "restricting ALL traffic to DROP\n");
			}
			else if (!strcmp(param[i+1], "pass")){
				cfg->restrict_all=PASS;
				aParse(conn, "restricting ALL traffic to PASS\n");
			}
			else aParse(conn, "restrict ALL traffic command invalid!\n");
		}
		else if (!strcmp(param[i], "local")) {
			if (!strcmp(param[i+1], "drop")){
				cfg->restrict_local=DROP;
				aParse(conn, "restricting LOCAL traffic to DROP\n");
			}
			else if (!strcmp(param[i+1], "pass")){
				cfg->restrict_local=PASS;
				aParse(conn, "restricting LOCAL traffic to PASS\n");
			}
			else aParse(conn, "restrict LOCAL traffic command invalid!\n");
		}

		else aParse(conn, "restrict command '%s' invalid!\n", param[i]);

		i+=2;
	}
	FW_CHECK_CHANGED(time(NULL))
	return PARSE_OK;
}
//////////////////////////////////////////////////////////////////////////////////////////
void sPMessagesGenerator(ServiceProcessor_cfg *cfg) {
	u_char fill=0;
	struct time_counters tc;
	time_t now;
	struct timeval start, stop;
	unsigned long len;

	now=time(NULL);
	gettimeofday(&start, NULL);
	
	if (unsigned(now - aGetActual('H',now)) <= 2*cfg->delay) {
		fill=1;
		PrepareTimeCounters(&tc);
	}
	
	netams_rwlock_rdlock(&Units->rwlock);

	for (NetUnit *u=Units->root; u!=NULL; u=u->next) {
		if(!u->ap) continue; // no need to check unit without account policy
		netams_rwlock_wrlock(&u->ap->rwlock);
		for (policy_data *pd=u->ap->root; pd!=NULL; pd=pd->next) {
			if ( fill || (unsigned(now-pd->flow.from) >= cfg->lifetime) ) {
				if( pd->flow.in!=0 || pd->flow.out!=0 ) {
					//update counters for current flow
					len=pd->flow.in;
					pd->h.in+=len;
					pd->d.in+=len;
					pd->w.in+=len;
					pd->m.in+=len;

					len=pd->flow.out;
					pd->h.out+=len;
					pd->d.out+=len;
					pd->w.out+=len;
					pd->m.out+=len;
				
					pd->to=now;
					//construct messages for policy_data pd of unit u
					sPConstructStoreMessages(u->id,pd);
					
					pd->flow.in=pd->flow.out=0;
				}
				pd->flow.from=now;
				
				if (fill) FillTimeCounters(pd,&tc);
			}
		}
		netams_rwlock_unlock(&u->ap->rwlock);
	}
	netams_rwlock_unlock(&Units->rwlock);

	gettimeofday(&stop, NULL);

	aDebug(DEBUG_PROC_MUX, "lookup takes %.4f  seconds\n", ((double)(stop.tv_usec-start.tv_usec))/1000000);
}
//////////////////////////////////////////////////////////////////////////
void sPMessagesMultiplexer(ServiceProcessor_cfg *cfg) {
	Message *msg;
	Message_Store *smsg;
	Message_Read  *rmsg;
	u_char read_warning=0;

	storage_list *item, *root=NULL, *read_sum_processor=NULL;
	aStorageGetAcceptedAll(ST_CONN_SUMMARY, &root);
	aStorageGetAcceptedAll(ST_CONN_RAW, &root);

	for (item=root; item!=NULL; item=item->next) {
		if (item->cfg->accept[ST_CONN_SUMMARY]) {
			read_sum_processor=item;
			break;
		}
	}
	
	while((msg=Mux_in->TryPop())) {
		switch (msg->type) {
		case MSG_STORE:
			smsg=(Message_Store*)msg;
			
			switch (smsg->prefix){
				case 'M':
				case 'W':
				case 'D':
				case 'H':
				case 'O':
					//here we depends that item list ordered and SUMMARY goes first
					for (item=root; item!=NULL && item->cfg->accept[ST_CONN_SUMMARY]; item=item->next) {
						item->cfg->in->Push(msg);
						aDebug(DEBUG_PROC_MUX, "P->ST %c st:%u unit:%06X acct:%06X from:%lu to:%lu\n", smsg->prefix, item->s->instance, smsg->netunit, smsg->ap, smsg->data->from, smsg->ts);
					}
					break;
				case 'F':
					for (item=root; item!=NULL; item=item->next) {
						if (item->cfg->accept[ST_CONN_RAW]) {
							item->cfg->in->Push(msg);
							aDebug(DEBUG_PROC_MUX, "P->ST %c st:%u unit:%06X acct:%06X from:%lu to:%lu\n", smsg->prefix, item->s->instance, smsg->netunit, smsg->ap, smsg->data->from, smsg->ts);
						}
					}
					if (cfg->fifo)
						cfg->fifo->Push(msg); //pass F message to billing to be counted
					break;
			} //switch
			break;

		case MSG_READ:
			if (!read_sum_processor) {
				if(!read_warning) {
					read_warning=1;
					aLog(D_WARN, "Can not perform READ request: no storages registered\n");
				}
				break;
			}

			rmsg=(Message_Read*)msg;
			read_sum_processor->cfg->out->Push(msg);
			aDebug(DEBUG_PROC_MUX, "P<-ST H,D,W,M st:%u unit:%06X acct:%06X to now\n",
				read_sum_processor->s->instance, rmsg->netunit, rmsg->ap);
			break;
		default:
			aDebug(DEBUG_PROC_MUX, "out <%u> (unknown message type)\n", msg->type);
			break;
		} // this message processed
		
		//pop message from Mux
		//we need to use TryPop() before cause possible race 
		//when pushing to storages
		Mux_in->Pop();

	} // all messages processed
	
	//wakeup all storages
	for (item=root; root!=NULL; item=root) {
		item->s->Wakeup();
		root=item->next;
		aFree(item);
	}
}
//////////////////////////////////////////////////////////////////////////
void cMac(Connection *conn, char *param[]) {
	u_char mac_control_block=0, mac_control_alert=0, mac_control_fixate=0, mac_control_control=0;
	if (!strcasecmp(param[1], "control")){
		for (int i=2; param[i]!=empty; i++) {
			if (!strcasecmp(param[i], "alert")) mac_control_alert=1;
			else if (!strcasecmp(param[i], "block")) mac_control_block=1;
		}
		mac_control_control=1;
	} else if (!strcasecmp(param[1], "fixate")) mac_control_fixate=1;
	else { aParse(conn, "unknown MAC command\n"); return; }
	
	unsigned num_units;
	// make a snapshot of units [id:ip:mac] table
	netams_rwlock_rdlock(&Units->rwlock);
	num_units=Units->num_units;
	
	MacControlEntry *table = (MacControlEntry*)aMalloc(num_units * sizeof (struct MacControlEntry));
	NetUnit *d;
	int i;
	MacControlEntry *t;
	for(d=Units->root, i=0; d!=NULL; d=d->next, i++)	{
		t=&table[i]; //*sizeof(struct MacControlEntry);
		t->id = d->id;
		t->state = MAC_CTL_SKIP;
		t->sp = d->sys_policy;
		if (!t->sp || t->sp&SP_DENY_MAC) { // do check only if no syspolicy or syspolicy==SP_MAC
			if (d->type==NETUNIT_HOST) {
				NetUnit_host *h = (NetUnit_host *)d;
				memcpy(&t->ip, &h->ip, sizeof (struct in_addr));
				if (h->mac) {
					memcpy(&t->mac, h->mac, sizeof (struct ether_addr));
					t->state = MAC_CTL_CHECK;
				}
			}
			else if (d->type==NETUNIT_USER) {
				NetUnit_user *h = (NetUnit_user *)d;
				memcpy(&t->ip, &h->ip, sizeof (struct in_addr));
				if (h->mac) {
					memcpy(&t->mac, h->mac, sizeof (struct ether_addr));
					t->state = MAC_CTL_CHECK;
				}
			} // do check only if no syspolicy or syspolicy==SP_MAC
			if (mac_control_fixate) t->state = MAC_CTL_FIXATE;
		}
	}		
	netams_rwlock_unlock(&Units->rwlock);
	
	// cycle through system ARP cache and do check
	// pizjeno from our s_login.c
	#if defined(FREEBSD) || defined(OPENBSD)
		char *buffer;
		struct ether_addr *e;
		int mib[6];
		size_t needed;
		char *lim, *next;
		struct rt_msghdr *rtm;
		struct sockaddr_inarp *sin;
		struct sockaddr_dl *sdl;
	
		mib[0] = CTL_NET; mib[1] = PF_ROUTE; mib[2] = 0; mib[3] = AF_INET; mib[4] = NET_RT_FLAGS; mib[5] = RTF_LLINFO;
		if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0) return;
		if ((buffer = (char*)aMalloc(needed)) == NULL) return;
		if (sysctl(mib, 6, buffer, &needed, NULL, 0) < 0) return;
	
		lim = buffer + needed;
	
		for (next = buffer; next < lim; next += rtm->rtm_msglen) {
			rtm = (struct rt_msghdr *)next;
			sin = (struct sockaddr_inarp *)(rtm + 1);
			sdl = (struct sockaddr_dl *)((char *)sin + ROUNDUP(sin->sin_len));
			e=(struct ether_addr*)LLADDR(sdl);
			aMacControl_check(table, num_units, &sin->sin_addr, e);
		}
		aFree(buffer);
	#endif
	
	#if defined(LINUX)
		char *buffer;
		char b_ip[32], b_hw[32], b_trash[32];
		struct in_addr in;
		struct ether_addr *e;
		FILE *arp;
		
		arp=fopen("/proc/net/arp", "rt");
		if (arp==NULL) { aDebug(DEBUG_LOGIN, "failed to open /proc/net/arp\n"); return; }
		
		if ((buffer = (char*)aMalloc(256)) == NULL) return;

		if (NULL==fgets(buffer, 255, arp)) goto FINISH_LINUX_CHECK_P;
	
		while (!feof(arp)){
			if (NULL==fgets(buffer, 255, arp)) break;
			bzero(b_ip, 31); bzero(b_hw, 8);
			sscanf(buffer, "%s%s%s%s", b_ip, b_trash, b_trash, b_hw);
			if (inet_aton(b_ip, &in) && (e=ether_aton(b_hw)))
				aMacControl_check(table, num_units, &in, e);
		}
	
	FINISH_LINUX_CHECK_P:
		fclose(arp);
		aFree(buffer);
	#endif
	
	// check the table again for matched violators & enforce
	NetUnit *u;
	struct ether_addr **mac;
	unsigned mac_fixate_chk=0, mac_fixate_upd=0;
	char buf[32];

	//if (mac_control_fixate) netams_rwlock_wrlock(&Units->rwlock);
	for (unsigned i=0; i<num_units; i++){
		t=&table[i];//*sizeof(struct MacControlEntry);
		if (t->state!=MAC_CTL_SKIP) ProcessorCfg.mac_control_units_checked++;
		if (t->state==MAC_CTL_VIOL) {
			aLog(D_WARN, "MAC violated OID %06X IP %s ARP_MAC %s\n", t->id, inet_ntop(AF_INET, &t->ip, buf, 32), ether_ntoa(&t->mac));
			LogEvent(SYSTEM, t->id, 0, 0, "MAC FAIL OID %06X IP %s ARP_MAC %s\n", t->id, buf, ether_ntoa(&t->mac));
			ProcessorCfg.mac_control_units_violated++;
			u=Units->getUnitById(t->id);
			if (mac_control_block) u->SetSysPolicy(SP_DENY_MAC, ADD); 
			if (mac_control_alert) aMacControl_alert(t, u, 1); 
		}
		else if (t->state==MAC_CTL_BACK) {
			aLog(D_WARN, "MAC restored OID %06X IP %s ARP_MAC %s\n", t->id, inet_ntop(AF_INET, &t->ip, buf, 32), ether_ntoa(&t->mac));
			LogEvent(SYSTEM, t->id, 0, 0, "MAC RESTORED OID %06X IP %s ARP_MAC %s\n", t->id, buf, ether_ntoa(&t->mac));
			u=Units->getUnitById(t->id);
			if (mac_control_block) u->SetSysPolicy(SP_DENY_MAC, REMOVE); 
			if (mac_control_alert) aMacControl_alert(t, u, 0); 
		}
		else if (t->state==MAC_CTL_UPDATE) {
			mac_fixate_chk++;
			u=Units->getUnitById(t->id);
			if (u->type==NETUNIT_HOST) mac = &((NetUnit_host *)u)->mac;
			else if (u->type==NETUNIT_USER) mac = &((NetUnit_user *)u)->mac;
			else continue;
			if (*mac==NULL) {
				*mac=(struct ether_addr *)aMalloc(sizeof (struct ether_addr));
				mac_fixate_upd++;
			}
			memcpy(*mac, &t->mac, sizeof (struct ether_addr));
		}
	}	
	if (mac_control_fixate) {
		//netams_rwlock_unlock(&Units->rwlock);
		aParse(conn, "MAC fixate: checked %u(%u) units, updated %u\n", mac_fixate_chk, num_units, mac_fixate_upd);
	}
	aFree(table);
}
//////////////////////////////////////////////////////////////////////////
void aMacControl_check(MacControlEntry* table, unsigned num_units, in_addr *ip, struct ether_addr* mac){
	MacControlEntry *t;
	int check_flag;
	for (unsigned i=0; i<num_units; i++){
		t=&table[i]; //*sizeof(struct MacControlEntry);
		if (t->state==MAC_CTL_CHECK) { // this unit has some MAC to do checking
			if (!memcmp(&t->ip, ip, sizeof (struct in_addr))) { // IP matched
				check_flag=memcmp(&t->mac, mac, sizeof (struct ether_addr));
				//printf("ip=%s mac=%s state=%u sp=%d check=%d\n", inet_ntoa(*ip), ether_ntoa(mac), t->state, t->sp, check_flag);
				if (check_flag && !(t->sp&SP_DENY_MAC)) { // violated
					t->state=MAC_CTL_VIOL;
					//printf("\tviolated! %d\n", t->state);
					memcpy(&t->mac, mac, sizeof (struct ether_addr)); // put actual MAC (from ARP) onto the table
				} else if (!check_flag && t->sp&SP_DENY_MAC) { // restored back
					t->state=MAC_CTL_BACK;
				}
			}
		} else if (t->state==MAC_CTL_FIXATE) { // just copy MAC onto table
			if (!memcmp(&t->ip, ip, sizeof (struct in_addr))) { // IP matched
				memcpy(&t->mac, mac, sizeof (struct ether_addr));
				t->state=MAC_CTL_UPDATE;
			}
		}
	}
}
//////////////////////////////////////////////////////////////////////////
void aMacControl_alert(MacControlEntry* t, NetUnit *u, u_char dir){
	// pizjeno from our s_quota.c
	Service *ss=NULL;
	ss=Services->getServiceNextByType(SERVICE_ALERTER,ss);
	if (!ss) return;
	ServiceAlerter_cfg *alerter=(ServiceAlerter_cfg*)ss->cfg;
	if (!alerter->queue) return; // alerter service is not running

	Message *msg;
	alert *al;

	msg = MsgMgr->New(MSG_ALERT);
	
	al=((Message_Alert*)msg)->al;
	al->sent=time(NULL);
	al->expired=al->sent+60*60; // one hour expire
	al->report_id=0x06101;
	al->tries=0;
	al->alert_id=alerter->queue->total_items+1; //here

	// recipients are all users with UPERM_ALL
	netams_mutex_lock(&Users->lock);
	netams_mutex_lock(&Users->lock);
	User *d;
	u_char i=0;
	for(d=Users->root; d!=NULL && i<MAX_ID_PER_ALERT; d=d->next)	{
		if (d->permissions==UPERM_ALL && !d->hidden) {
			al->user_id[i]=d->id;
			aDebug(DEBUG_ALERT, "USER (mac control) RCPT %u added: %06X\n", i+1, d->id);
			i++;
		}
	}
	netams_mutex_unlock(&Users->lock);
	netams_mutex_unlock(&Users->lock);
	if (i==0) { 
		aDebug(DEBUG_ALERT, "alert (mac control) abandoned because of no recipients\n");
		return; // nowhere to send
	}
	
	char *subject, *message, buffer[255];
	subject=message=NULL;
	
	timeU2T(time(NULL), buffer);

	print_to_string(&message, "This is automatically generated report by %s, version %u.%u.%u (build %u)\nTime: %s\n\n", aaa_fw_software_name, aaa_fw_major_version, aaa_fw_minor_version, aaa_fw_subversion, aaa_fw_build_version, buffer);
	
	switch (dir) {
		case 0:
			print_to_string(&subject, "NeTAMS MAC Control: unit %s MAC returned", u->name?u->name:"<>");
			print_to_string(&message, "NeTAMS MAC Control service have detected hardware (MAC) address returned:\n");
			break;
		case 1:
			print_to_string(&subject, "NeTAMS MAC Control: unit %s MAC violation", u->name?u->name:"<>");
			print_to_string(&message, "NeTAMS MAC Control service have detected hardware (MAC) address violation:\n");
			break;
	}

	print_to_string(&message, "\nOID:\t\t\t%06X\nName:\t\t\t%s\nIP:\t\t\t%s\nARP MAC:\t\t%s\n", u->id, u->name?u->name:"<\?\?>", inet_ntop(AF_INET, &t->ip, buffer, 32), ether_ntoa(&t->mac));
	//kill(getpid(), 17);
	if (u->type==NETUNIT_HOST) {
		NetUnit_host* h = (NetUnit_host*)u;
		if (h->mac) print_to_string(&message, "Allowed MAC:\t%s\n", ether_ntoa(h->mac));	
	}
	else if (u->type==NETUNIT_USER) {
		NetUnit_user* h = (NetUnit_user*)u;
		if (h->mac) print_to_string(&message, "Allowed MAC:\t%s\n", ether_ntoa(h->mac));	
	}
		
	al->data=NULL;
	print_to_string(&al->data, "%s\n%s\n", subject, message);
	aFree(subject); aFree(message);
	
	aDebug(DEBUG_ALERT, "alert (mac control) %u complete, data is %u bytes\n", al->alert_id, strlen(al->data));
	al->alert_id=(unsigned)alerter->queue->Push(msg);
	alerter->self->Wakeup();

}
//////////////////////////////////////////////////////////////////////////
