/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: s_storage.c,v 1.84.2.8 2006/03/30 11:34:58 jura Exp $ */

#include "netams.h"

//////////////////////////////////////////////////////////////////////////
void *sStorage(void *s);
void sStorageCancel(void *);
unsigned stRead(Service *s);
void stStore(Service *s);
//////////////////////////////////////////////////////////////////////////
void sStorageInit(Service *s){
	s->cfg = (ServiceStorage_cfg*)aMalloc(sizeof(ServiceStorage_cfg));

	ServiceStorage_cfg *cfg=(ServiceStorage_cfg*)s->cfg;
	cfg->type=UNKNOWN;
	cfg->hostname=cfg->username=cfg->password=cfg->dbname=cfg->socket=NULL;
	cfg->port=0;
	cfg->in=new FIFO(MAX_UNITS*5);
	cfg->out=new FIFO(MAX_UNITS*5);
	cfg->oids=cfg->raw=cfg->summary=NULL;
	cfg->db_path=NULL;
	cfg->retry=ST_RADIUS_DEFAULT_RETRY; cfg->timeout=ST_RADIUS_DEFAULT_TIMEOUT; 
	cfg->pid = getpid(); 
	cfg->nas_ip.s_addr=INADDR_ANY;	
	bzero(cfg->fd, ST_CONN_TYPES_NUM*sizeof(void*));
	for (int i=0; i<ST_CONN_TYPES_NUM; i++) cfg->accept[i]=1; // by default, each storage accepts everything

	pthread_create(&(s->t_id), NULL, &sStorage, s);
}

//////////////////////////////////////////////////////////////////////////
void sStorageProcessCfg(char *param[], Connection *conn, u_char no_flag){
	ServiceStorage_cfg *cfg=(ServiceStorage_cfg*)conn->service->cfg;

	if (!strcasecmp(param[0], "type")) {
		if(0) ;
#if defined(USE_HASH)
		else if (!strcasecmp(param[1], "hash")) {
			aParse(conn, "storage type is set to HASH\n");
			cfg->type=HASH;
		}
#endif
#if defined(USE_MYSQL)
		else if (!strcasecmp(param[1], "mysql")) {
			aParse(conn, "storage type is set to MySQL\n");
			cfg->type=MY_SQL;
		}
#endif
#if defined(USE_POSTGRES)
		else if (!strcasecmp(param[1], "postgres")) {
			aParse(conn, "storage type is set to Postgres\n");
			cfg->type=POSTGRES;
		}
#endif
#if defined(USE_ORACLE)
		else if (!strcasecmp(param[1], "oracle")) {
			aParse(conn, "storage type is set to Oracle\n");
			cfg->type=ORACLE;
		}
#endif
#if defined(USE_LIBRADIUS)
		else if (!strcasecmp(param[1], "radius")) {
			aParse(conn, "storage type is set to Radius\n");
			cfg->type=LIBRADIUS;
		}
#endif
		else {
			aParse(conn, "storage type is unknown\n");
			cfg->type=UNKNOWN;
		}
	}		
	else if (!strcasecmp(param[0], "user")) {
		if (param[1]!=empty &&
            (cfg->type==MY_SQL || cfg->type==POSTGRES || cfg->type == ORACLE)) {
			if(cfg->username) aFree(cfg->username);
			cfg->username=set_string(param[1]);
			aParse(conn, "username is set to %s\n", cfg->username);
		} 
		else aParse(conn, "cannot set username for this storage type\n");
	}
	else if (!strcasecmp(param[0], "path")) {
		if (param[1]!=empty && cfg->type==HASH ) {
			if(cfg->username) aFree(cfg->username);
			cfg->username=set_string(param[1]);
			aParse(conn, "path is set to %s\n", cfg->username);
		} 
		else aParse(conn, "cannot set path for this storage type\n");
	}
	else if (!strcasecmp(param[0], "password")) {
		if (param[1]!=empty &&
            (cfg->type==MY_SQL || cfg->type==POSTGRES || cfg->type == ORACLE || cfg->type == LIBRADIUS) ) {
			if(cfg->password) aFree(cfg->password);
			cfg->password=set_string(param[1]);
			aParse(conn, "password is set to %s\n", cfg->password);
		} 
		else aParse(conn, "cannot set password for this storage type\n");
	}
	else if (!strcasecmp(param[0], "host")) {
		if (param[1]!=empty && (cfg->type==MY_SQL || cfg->type==POSTGRES || cfg->type == LIBRADIUS)) {
			if(cfg->hostname) aFree(cfg->hostname);
			cfg->hostname=set_string(param[1]);
			aParse(conn, "hostname is set to %s\n", cfg->hostname);
		} 
		else aParse(conn, "cannot set hostname for this storage type\n");
	}
	else if (!strcasecmp(param[0], "dbname")) {
		if (param[1]!=empty &&
            (cfg->type==MY_SQL || cfg->type==POSTGRES || cfg->type == ORACLE) ) {
			if(cfg->dbname) aFree(cfg->dbname);
			cfg->dbname=set_string(param[1]);
			aParse(conn, "dbname is set to %s\n", cfg->dbname);
		} 
		else aParse(conn, "cannot set dbname for this storage type\n");
	}
	else if (!strcasecmp(param[0], "socket")) {
		if (param[1]!=empty && (cfg->type==MY_SQL || cfg->type==POSTGRES) ) {
			if(cfg->socket) aFree(cfg->socket);
			cfg->socket=set_string(param[1]);
			aParse(conn, "socket is set to %s\n", cfg->socket);
		}
		else aParse(conn, "cannot set socket for this storage type\n");
	}
	 else if (!strcasecmp(param[0], "port")) {
		if (param[1]!=empty && (cfg->type==MY_SQL || cfg->type==POSTGRES || cfg->type == LIBRADIUS) ) {
			cfg->port=strtol(param[1], NULL, 10);
			aParse(conn, "port is set to %u\n", cfg->port);
		}
		else aParse(conn, "cannot set port for this storage type\n");
	}
#if defined(USE_LIBRADIUS)
	 else if (!strcasecmp(param[0], "retry")) {
		if (param[1]!=empty && cfg->type == LIBRADIUS) {
			cfg->retry=strtol(param[1], NULL, 10);
			aParse(conn, "radius packet retry count is set to %u\n", cfg->retry);
		}
		else aParse(conn, "cannot set packet retry count for this storage type\n");
	}
	 else if (!strcasecmp(param[0], "timeout")) {
		if (param[1]!=empty && cfg->type == LIBRADIUS) {
			cfg->timeout=strtol(param[1], NULL, 10);
			aParse(conn, "radius packet timeout is set to %u\n", cfg->timeout);
		}
		else aParse(conn, "cannot set packet timeout for this storage type\n");
	 }
	 else if (!strcasecmp(param[0], "nas-ip")) {
		if (param[1]!=empty && cfg->type == LIBRADIUS) {
			if (inet_aton(param[1], &(cfg->nas_ip)))
				aParse(conn, "radius nas-ip is set to %s\n", param[1]);
			else cfg->nas_ip.s_addr=INADDR_ANY;
		}
		else aParse(conn, "cannot set packet nas-ip for this storage type\n");
	}
#endif	
	 else if (!strcasecmp(param[0], "accept")) {
	 	for (int i=0; i<ST_CONN_TYPES_NUM; i++) cfg->accept[i]=0; // if accept is specified, flush everything
	 	int except=1;
	 	for (int j=1; param[j]!=empty; j++ ){
 			if (!strcasecmp(param[j], "except")) {
 				except=0;
 			}
 			else if (!strcasecmp(param[j], "all")) {
 				for (int i=0; i<ST_CONN_TYPES_NUM; i++) cfg->accept[i]=except;
 			}
 			else {
 				for (int i=1; i<ST_CONN_TYPES_NUM; i++) {
 					if (!strcasecmp(st_table_name[i], param[j])) cfg->accept[i]=except;
 				}
 			}
	 	}
	 	char *tmp=NULL;
	 	for (int i=1; i<ST_CONN_TYPES_NUM; i++) if (cfg->accept[i]==1) print_to_string(&tmp, " %s", st_table_name[i]);
	 	aParse(conn, "this storage will accept following data:%s\n", tmp);
	 	aFree(tmp);
	}
	else aParse(conn, "unknown storage command: %s\n", param[0]);
}
//////////////////////////////////////////////////////////////////////////
void *sStorage(void *ss){
	Service *s=(Service*)ss;

	aLog(D_INFO, "service storage:%u thread started (%p)\n", s->instance, s->cfg);
	pthread_cleanup_push(sStorageCancel, s);
	SET_CANCEL();
	s->t_id=pthread_self();

	ServiceStorage_cfg *cfg=(ServiceStorage_cfg*)s->cfg;
	
	// now we should sleep before starting storage service exec
	if(!(s->flags&SERVICE_RUN)) s->Sleep();

	switch (cfg->type) {
		case HASH:
		#ifdef USE_HASH
			{
				char hash_path[255];
				char *c;

				if (cfg->username) {
                			strcpy(hash_path, cfg->username);
				} else {
					c=strrchr(config_file_name, '/');
					if (c)
						strncpy(hash_path, config_file_name, (c-config_file_name));
					else
						strcpy(hash_path, "./");
				}
				print_to_string(&cfg->raw, "%s/netams-raw.db", hash_path);
				print_to_string(&cfg->summary, "%s/netams-sumary.db", hash_path);
			}
			aLog(D_INFO, "storage:%d working with HASH\n", s->instance);
				
			while (1) {
				//process input queue
				if(stLoadHash(s)) {
					//process output queue
					stSaveHash(s);
				}

				//this is cancel point
				s->Sleep();
			}
		#else 
			aLog(D_WARN, "storage type HASH is not compiled in!\n");
		#endif
			break;
		case POSTGRES:
		case MY_SQL:
		case ORACLE:
#if defined(USE_MYSQL) || defined(USE_POSTGRES) || defined(USE_ORACLE)
			print_to_string(&cfg->oids, "storage.%u.oids", s->instance);
			print_to_string(&cfg->raw, "storage.%u.raw", s->instance);
			print_to_string(&cfg->summary, "storage.%u.summary", s->instance);

			aLog(D_INFO, "storage:%d working with SQL\n", s->instance);
			
			//write previous data to database
			//there is guarantee this data will be written, at least in next writes
			//but no, that it will be counted propely if no database connection available
			stSaveSql(s,cfg->summary,ST_CONN_SUMMARY);

			while (1) {
				//process input queue
				if(stRead(s)) {
					//process output queue
					stStore(s);
				}
				s->Sleep();
			}
		#else	
			aLog(D_WARN, "storage type SQL is not compiled in!\n");
		#endif
			break;
		case LIBRADIUS:
			#ifdef USE_LIBRADIUS
			aLog(D_INFO, "storage:%d working with Radius\n", s->instance);
			stOpenRadius(s);
			while(1) {
				stStoreRadius(s);
				s->Sleep();
			}
			#else	
				aLog(D_WARN, "storage type RADIUS is not compiled in!\n");
			#endif
		break;

		case UNKNOWN:
		default:
			aLog(D_WARN, "storage type undefined, service will be finished\n");
			break;
	}

	pthread_cleanup_pop(1);
	
	//cleanup if service can't be started
	Services->Delete(s);
	delete s;

	return NULL;
}

//////////////////////////////////////////////////////////////////////////
void sStorageCancel(void *v){
	Service *s = (Service*)v;
	ServiceStorage_cfg *cfg=(ServiceStorage_cfg*)s->cfg;

	aLog(D_INFO, "cancelling service storage:%u, flushing %u messages\n", s->instance, cfg->in->num_items);
	switch (cfg->type) {
	#ifdef USE_HASH
		case HASH:
			if(stLoadHash(s)) stSaveHash(s);
			break;
	#endif
#if defined(USE_MYSQL) || defined(USE_POSTGRES) || defined(USE_ORACLE)
		case POSTGRES:
		case MY_SQL:
        case ORACLE:
			if(stRead(s)) stStore(s);
			break;
	#endif
	#ifdef USE_LIBRADIUS
		case LIBRADIUS:
			stStoreRadius(s);
			break;
	#endif
		default: break;
	}

	for(u_char type=0;type<ST_CONN_TYPES_NUM;type++)
		if(cfg->fd[type]) stCloseSql(s,(st_conn_type)type);

	delete cfg->in;
	delete cfg->out;
	if(cfg->username) aFree(cfg->username);
	if(cfg->password) aFree(cfg->password);
	if(cfg->dbname)   aFree(cfg->dbname);
	if(cfg->db_path)  aFree(cfg->db_path);
	if(cfg->hostname) aFree(cfg->hostname);
	if(cfg->socket) aFree(cfg->socket);
	if(cfg->oids) 	aFree(cfg->oids);
	if(cfg->raw) 	aFree(cfg->raw);
	if(cfg->summary) aFree(cfg->summary);
}

//////////////////////////////////////////////////////////////////////////
void sStorageListCfg(Service *s, FILE *f, u_char flags){
	ServiceStorage_cfg *cfg=(ServiceStorage_cfg*)s->cfg;

	switch (cfg->type) {
		case HASH:
			fprintf(f, "type hash\n"); 
			if (cfg->username) fprintf(f, "path %s\n", cfg->username); 
			break;
		case MY_SQL: 		
			fprintf(f, "type mysql\n"); 
			if (cfg->socket) fprintf(f, "socket %s\n", cfg->socket);
			if (cfg->hostname) fprintf(f, "host %s\n", cfg->hostname); 
			if (cfg->port && cfg->port!=3306) fprintf(f, "port %u\n", cfg->port);
			if (cfg->username) fprintf(f, "user %s\n", (flags&CFG_NO_PASSWORDS)?"***":cfg->username); 
			if (cfg->password) fprintf(f, "password %s\n", (flags&CFG_NO_PASSWORDS)?"***":cfg->password); 
			if (cfg->dbname) fprintf(f, "dbname %s\n", cfg->dbname); 
			break;
		case POSTGRES: 
			fprintf(f, "type postgres\n");
			if (cfg->socket) fprintf(f, "socket %s\n", cfg->socket);
			if (cfg->hostname) fprintf(f, "host %s\n", cfg->hostname);
			if (cfg->port && cfg->port!=5432) fprintf(f, "port %u\n", cfg->port);
			if (cfg->username) fprintf(f, "user %s\n", (flags&CFG_NO_PASSWORDS)?"***":cfg->username); 
			if (cfg->password) fprintf(f, "password %s\n", (flags&CFG_NO_PASSWORDS)?"***":cfg->password);
			if (cfg->dbname) fprintf(f, "dbname %s\n", cfg->dbname); 
			break;
		case ORACLE:
			fprintf(f, "type oracle\n"); 
			if (cfg->username) fprintf(f, "user %s\n", (flags&CFG_NO_PASSWORDS)?"***":cfg->username); 
			if (cfg->password) fprintf(f, "password %s\n", (flags&CFG_NO_PASSWORDS)?"***":cfg->password);
			if (cfg->dbname) fprintf(f, "dbname %s\n", cfg->dbname); 
			break;
		case LIBRADIUS:
			fprintf(f, "type radius\n"); 
			if (cfg->hostname) fprintf(f, "host %s\n", cfg->hostname);
			if (cfg->password) fprintf(f, "password %s\n", (flags&CFG_NO_PASSWORDS)?"***":cfg->password);
			if (cfg->port && cfg->port!=1813) fprintf(f, "port %u\n", cfg->port);
			if (cfg->timeout!=ST_RADIUS_DEFAULT_TIMEOUT) fprintf(f, "timeout %u\n", cfg->timeout);
			if (cfg->nas_ip.s_addr!=INADDR_ANY) {
				char *name = (char*)aMalloc(256);
				inet_ntop(AF_INET, &(cfg->nas_ip), name, 256);
				fprintf(f, "nas-ip %s\n", name);
				aFree(name);
				}
			if (cfg->retry!=ST_RADIUS_DEFAULT_RETRY) fprintf(f, "retry %u\n", cfg->retry);
			break;
		default:
			break;
	}
	if (cfg->accept[ST_CONN_UNDEF]==1) {
		fprintf(f, "accept all");
		int except_printed=0;
		for (int i=1; i<ST_CONN_TYPES_NUM; i++) if (cfg->accept[i]==0) {
			if (except_printed==0) { except_printed=1; fprintf(f, " except"); }
			fprintf(f, " %s", st_table_name[i]);
		}
	} else {
		fprintf(f, "accept");
		for (int i=1; i<ST_CONN_TYPES_NUM; i++) if (cfg->accept[i]==1) {
			fprintf(f, " %s", st_table_name[i]);
		}
	}
	fprintf(f, "\n"); // accept ...
		
	fprintf(f, "\n"); // finish the service config
}
//////////////////////////////////////////////////////////////////////////
unsigned stRead(Service *s) {
	ServiceStorage_cfg *cfg=(ServiceStorage_cfg*)s->cfg;
	Message *msg;
	Message_Read *rmsg;
	unsigned res=0;
	NetUnit *u;
	void *fd;

	if(!cfg->out->TryPop()) return 1;
		
	if(!(fd=stOpenSql(s,ST_CONN_SUMMARY))) {
		return 0;
	}

	while((msg=cfg->out->TryPop())) {
		if (msg->type==MSG_READ) {
			rmsg=(Message_Read*)msg;
			
			if(rmsg->pdata->flags&POLICY_DATA_UNUSED) {
				aFree(rmsg->pdata);
				goto CLEAN;
			}
			if(!(u=Units->getUnitById(rmsg->netunit))) {
				rmsg->pdata->flags&=~POLICY_DATA_LOCKED; //unlock pdata
				goto CLEAN;
			}
			aDebug(DEBUG_STORAGE, "ST<-SQL/H,D,W,M rd st:%u unit:%06X acct:%06X\n", 
				s->instance, rmsg->netunit, rmsg->ap);
			
			if(!u->ap) goto CLEAN; //somehow unit do not have account data

			//protect counters
			netams_rwlock_wrlock(&u->ap->rwlock);
			res=stLoadSql(s,fd,rmsg);
			rmsg->pdata->flags&=~POLICY_DATA_LOCKED; //unlock pdata
			netams_rwlock_unlock(&u->ap->rwlock);
			
			if(!res) return 0; // we cannot read

		} else {
			aDebug(DEBUG_STORAGE, "message type %d not READ - discarded\n",msg->type);
		}
CLEAN:
		msg=cfg->out->Pop();
	}
	
	return res;
}
//////////////////////////////////////////////////////////////////////////
void stStore(Service *s) {
	ServiceStorage_cfg *cfg=(ServiceStorage_cfg*)s->cfg;
	Message *msg;
	Message_Store *smsg;
	FILE *raw = NULL;
	FILE *summary = NULL;
	FILE *oids=NULL;
	
	unsigned num_raw=0;
	unsigned num_summary=0;
	unsigned num_oids=0;
	NetUnit *u;

	while((msg=cfg->in->TryPop())) {
		if (msg->type==MSG_STORE) {
			smsg=(Message_Store*)msg;
			aDebug(DEBUG_STORAGE, "ST->SQL/%c wr st:%u unit:%06X acct:%06X from:%u to:%u\n", smsg->prefix, s->instance, smsg->netunit, smsg->ap, smsg->data->from, smsg->ts);
			switch (smsg->prefix){
				case 'F':
					if(!raw) {
						raw=fopen(cfg->raw,"a");
						if(!raw) {
							aLog(D_ERR, "Can't create temporary file %s: %s\n", cfg->raw, strerror(errno));
							return;
						}
						setlinebuf(raw);
					}
					fprintf(raw, "%u,%u,%lu,%lu,%llu,%llu\n", smsg->netunit, smsg->ap, smsg->data->from, smsg->ts, smsg->data->in, smsg->data->out);
					num_raw++;
					break;
				case 'O':
					if(!oids) {
						oids=fopen(cfg->oids,"a");
						if(!oids) {
							aLog(D_ERR, "Can't create temporary file %s: %s\n", cfg->oids, strerror(errno));
							return;
						}
						setlinebuf(oids);
					}
					u =Units->getUnitById(smsg->netunit);  
					if (u) { 
						fprintf(oids, "%u,%s\n", smsg->netunit, u->name);
						num_oids++;
					}
					break;
				case 'H':
				case 'D':
				case 'W':
				case 'M':
					if(!summary) {
						summary=fopen(cfg->summary,"a");
						if(!summary) {
							aLog(D_ERR, "Can't create temporary file %s: %s\n", cfg->summary, strerror(errno));
							return;
						}
						setlinebuf(summary);
					}
					fprintf(summary, "%c,%u,%u,%lu,%llu,%llu\n", smsg->prefix, smsg->netunit, smsg->ap, smsg->data->from, smsg->data->in, smsg->data->out);
					num_summary++;
					break;
			}
		} else {
			aDebug(DEBUG_STORAGE, "message type %d not STORE - discarded\n",msg->type);
 		}
		msg=cfg->in->Pop();
	}

	if(raw) {
		fclose(raw);
		aDebug(DEBUG_STORAGE, "SQL->HDD/raw %u queries\n", num_raw);
		stSaveSql(s,cfg->raw,ST_CONN_RAW);
	}	
	if(summary) {
		fclose(summary);
		aDebug(DEBUG_STORAGE, "SQL->HDD/summary %u queries\n", num_summary);
		stSaveSql(s,cfg->summary,ST_CONN_SUMMARY);
	}

	if(oids) {
		fclose(oids);
		aDebug(DEBUG_STORAGE, "SQL->HDD/oids %u queries\n", num_oids);
		stSaveSql(s,cfg->oids,ST_CONN_OIDS);
	}

}
//////////////////////////////////////////////////////////////////////////
Service *aStorageGetAccepted(st_conn_type type){
	Service *s=NULL;
	ServiceStorage_cfg *cfg;

	while((s = Services->getServiceNextByType(SERVICE_STORAGE, s))) {
		cfg = (ServiceStorage_cfg*)s->cfg;
//		aLog(D_INFO, "-- service type:%d, id=%d, type=%d, accept[]=%d\n", s->type, s->instance, type, cfg->accept[type]);
		if (cfg->accept[type]==1) return s;
	}
	return NULL;
}
//////////////////////////////////////////////////////////////////////////
void aStorageGetAcceptedAll(st_conn_type type, storage_list **list){
	Service *s=NULL;
	ServiceStorage_cfg *cfg;
	storage_list *sl, *t, *lst;

	while((s = Services->getServiceNextByType(SERVICE_STORAGE, s))) {
		cfg = (ServiceStorage_cfg*)s->cfg;
		if (cfg->accept[type]==1) {
			for (sl=*list, t=NULL, lst=NULL; sl!=NULL; sl=sl->next) {
				lst=sl;
				if (sl->s==s) { t=sl; break; }
			}
			if (!t) {
				sl = (storage_list*)aMalloc(sizeof (struct storage_list));
				sl->cfg=cfg;
				sl->s=s;
				sl->next=NULL;

				if (*list==NULL) { *list=sl; }
				else { lst->next=sl; }
			}
		}
	}
}
//////////////////////////////////////////////////////////////////////////

