/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: security.c,v 1.25 2005/09/08 13:36:30 anton Exp $ */

#include "netams.h"

/////////////////////////////////////////////////////////////////////////////////////////
// Permissions Management ///////////////////////////////////////////////////////////////////

char *permissions_list[256]; 

void aPermissionsInit(){

	for (unsigned i=0; i<=255; i++) permissions_list[i]=NULL;
	permissions_list[UPERM_NONE]="none";              
	permissions_list[UPERM_LOGIN_WEB]="weblogin";          
	permissions_list[UPERM_SHOWSTAT_HIS]="showstat_his";
	permissions_list[UPERM_SHOWOBJ_HIS]="showobj_his";      
	permissions_list[UPERM_MODIFYOBJ_HIS]="modifyobj_his";    
	permissions_list[UPERM_SHOWSTAT_ALL]="showstat";
	permissions_list[UPERM_SHOWOBJ_ALL]="showobj";          
	permissions_list[UPERM_SHOWCONFIG]="showconfig";       
	permissions_list[UPERM_RADIUS]="radius";       
	permissions_list[UPERM_MODIFYOBJ_ALL]="modifyobj";
	permissions_list[UPERM_MODIFYCONFIG]="modifyconfig";     
	permissions_list[UPERM_LOGIN_INT]="intlogin";        
	permissions_list[UPERM_SYSACTION]="sysaction";
	permissions_list[UPERM_ALL]="all";

#ifdef FREEBSD
#ifndef NETBSD
	crypt_set_format("md5");
#endif
#endif
}

u_char aCheckPerm(Connection *conn, u_char perm){
	if (conn->permissions < perm) {
		aLog(D_WARN, "connection %s: not enough permissions\n", conn->name);
		return 1;
		}
	return 0;
}

int aAuth(char *login, char *password){
	User *u;
	
	if (!login || !password) return -1;
	
	u=Users->getUser(login);
	
	if (!u) return -1;
	
	if (u->password) {
		if (strlen(password)==strlen(u->password)) 
			if (!strcmp(password, u->password)) return u->permissions;
	}

	if (u->crypted) {
		char *c=crypt(password, "$1$");
		if (strlen(c)==strlen(u->crypted)) 
			if (!strcmp(c, u->crypted)) return u->permissions;
	}

	return -1;
}

int cRadius(char *param[], Connection *conn){
	
	char *login=NULL, *password=NULL, *nas_id=NULL, *callback_id=NULL; 
	char *options=NULL;
	char buffer[32];
	
	if (strcasecmp(param[1], "auth")==0 && param[2]!=empty) {

		u_char i=3;
		while (param[i]!=empty) {
	 		if (strcasecmp(param[i], "login")==0) { 
				login=set_string(param[i+1]);
			} else if (strcasecmp(param[i], "password")==0) {
				password=set_string(param[i+1]);
			} else if (strcasecmp(param[i], "nas-id")==0) {
				nas_id=set_string(param[i+1]);
			} else if (strcasecmp(param[i], "callback-id")==0) {
				callback_id=set_string(param[i+1]);
	 		} else i--;
			i=i+2;
		}

		if (strcasecmp(param[2], "nas")==0 && login) {
			NetUnit *u = Units->getUnit(login);
			int fail=0; // can fail during callback_id checking
			if (u) {
				if ( (u->password==NULL && (password==NULL || password==empty) ) ||
					(u->password!=NULL && password!=NULL && strcmp(u->password, password)==0) ||
					(u->password!=NULL && password==NULL ) ) {
						int num_options=0;
						if (u->type==NETUNIT_USER) {
							NetUnit_user *us = (NetUnit_user*)u;
							if (us->ip.s_addr!=0) {
								num_options++;
								print_to_string(&options, "Framed-IP-Address: %s\n", inet_ntop(AF_INET, &(us->ip), buffer, 32));
							}
							if (callback_id)
								if (us->mac) 
									if (memcmp(ether_aton(callback_id), us->mac, sizeof(ether_addr))) fail=1;
						}
						else if (u->type==NETUNIT_HOST) {
							NetUnit_host *uh = (NetUnit_host*)u;
							if (uh->ip.s_addr!=0) {
								num_options++;
								print_to_string(&options, "Framed-IP-Address: %s\n", inet_ntop(AF_INET, &(uh->ip), buffer, 32));
							}
							if (callback_id)
								if (uh->mac) 
									if (memcmp(ether_aton(callback_id), uh->mac, sizeof(ether_addr))) fail=1;
						}
						if (u->sys_policy != SP_NONE) fail = 1; /* http://www.netams.com/ubb/cgi-bin/ultimatebb.cgi?ubb=get_topic&f=3&t=000155 */
						if (u->fp!=NULL) {
							policy_data *cpd;
							for (cpd=u->fp->root; cpd!=NULL; cpd=cpd->next) {
								num_options++;
								print_to_string(&options, "Filter-ID: %06X %s\n", cpd->policy->id, cpd->policy->name?cpd->policy->name:"<\?\?>");
							}
						}
						if (u->password!=NULL && password==NULL) {
							num_options++;
							print_to_string(&options, "User-Password: %s\n", u->password);
						}
						// add Session-Timeout handler here
						
						if (fail) {
							fprintf(conn->stream_w, "0 callback check failed for %s\n", login);
							LogEvent(RADIUS, u->id, 0, 0, "FAIL callback check for %s, nas-id %s callback-id %s", login, nas_id?nas_id:"<none>", callback_id?callback_id:"<none>");
							aLog(D_INFO, "RADIUS FAIL callback check for %s, nas-id %s callback-id %s\n", login, nas_id?nas_id:"<none>", callback_id?callback_id:"<none>");
						} else {
							fprintf(conn->stream_w, "1 %d\n%s", num_options, options);
							LogEvent(RADIUS, u->id, 0, 0, "SUCCESS login %s nas-id %s callback-id %s", login, nas_id?nas_id:"<none>", callback_id?callback_id:"<none>");
							aLog(D_INFO, "RADIUS SUCCESS login %s nas-id %s callback-id %s\n", login, nas_id?nas_id:"<none>", callback_id?callback_id:"<none>");
						}
					} else { 
						fprintf(conn->stream_w, "0 password incorrect for %s\n", login);
						LogEvent(RADIUS, u->id, 0, 0, "FAIL password incorrect for %s, nas-id %s callback-id %s", login, nas_id?nas_id:"<none>", callback_id?callback_id:"<none>");
						aLog(D_INFO, "RADIUS FAIL password incorrect for %s, nas-id %s callback-id %s\n", login, nas_id?nas_id:"<none>", callback_id?callback_id:"<none>");
					}
			} else  {
				fprintf(conn->stream_w, "0 login unknown: %s\n", login);
				LogEvent(RADIUS, 0, 0, 0, "FAIL login unknown: %s", login);
				aLog(D_INFO, "RADIUS FAIL login unknown: %s\n", login);
			}
		}
		else if (strcasecmp(param[2], "web")==0 && login) {
			NetUnit *u = Units->getUnit(login);
			User *uu = Users->getUser(login);
			if (u) {
				if ( (u->password==NULL && (password==NULL || password==empty) ) ||
					(u->password!=NULL && password!=NULL && strcmp(u->password, password)==0) ) {
						fprintf(conn->stream_w, "1 0 pass unit %s\n", login);
					} else fprintf(conn->stream_w, "0 web authorization password incorrect for unit: %s\n", login);
			}
			else if (uu) {
				if ( aAuth(login, password)!=-1) {
					if (!uu->hidden && uu->permissions >= UPERM_SHOWSTAT_HIS) fprintf(conn->stream_w, "1 0 pass user %s\n", login);
					else fprintf(conn->stream_w, "0 user %s insufficient rights\n", login);
				}
				else fprintf(conn->stream_w, "0 web authorization password incorrect for user: %s\n", login);
			}
			else fprintf(conn->stream_w, "0 web authorization login is unknown: %s\n", login);
		}
		else fprintf(conn->stream_w, "0 radius authorization source is unknown: %s\n", param[2]);
	} 
	else if (strcasecmp(param[1], "accnt")==0) {
		fprintf(conn->stream_w, "0 radius accounting is not implemented\n"); 
	} 
	else fprintf(conn->stream_w, "0 radius method unknown: %s\n", param[1]); 

	aFree(login); aFree(password); aFree(nas_id); aFree(callback_id); aFree(options);
	return PARSE_OK;
}
