/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: st_radius.c,v 1.15.2.3 2005/11/19 21:31:07 jura Exp $ */

#ifdef USE_LIBRADIUS

#include "netams.h"
#include "../lib/radlib/radlib.h"
#include "../lib/radlib/radlib_vs.h"

/*  this module will take RAW data messages from processor service, 
 *  and send it to radius server, accounting packets
 *  to specifiy radius server, we have to set hostname, password and 
 *  port parameters. default is "localhost, secret, 1813". username and dbname
 *  has no effect. other message types are not processed.
 */

int	radSendAccounting(Message_Store *, ServiceStorage_cfg *);

//////////////////////////////////////////////////////////////////////////
void stOpenRadius(Service *s){
	ServiceStorage_cfg *cfg=(ServiceStorage_cfg*)s->cfg;
	char *name = (char*)aMalloc(256);
	if (cfg->nas_ip.s_addr==INADDR_ANY) {
		// we don't have NAS IP specified, will try to obtain it from gethostbyname
		if (gethostname(name, 256)!=0) { aLog(D_WARN, "gethostname: %d (%s)\n", h_errno, strerror(h_errno)); return; }
	
		struct hostent *hp;
		hp = gethostbyname(name);
		if (hp==NULL) { herror("gethostbyname:"); aLog(D_ERR, "gethostbyname: %d (%s)\n", h_errno, strerror(h_errno)); return; }
		
		memcpy((char*)&cfg->nas_ip.s_addr, hp->h_addr_list[0], hp->h_length);
	}
	inet_ntop(AF_INET, &(cfg->nas_ip), name, 256);
	aDebug(DEBUG_STORAGE, "RADIUS: Our NAS-IP-Address will be %s\n", name);
	aFree(name);
}
//////////////////////////////////////////////////////////////////////////
void stStoreRadius(Service *s){
	
	ServiceStorage_cfg *cfg=(ServiceStorage_cfg*)s->cfg;
	Message *msg;
	Message_Store *smsg;
	unsigned num_raw=0;
	if (is_running==0 && global_return_code!=0) { cfg->timeout=0, cfg->retry=0; } // if we are finishing, speed up purge
		
	while((msg=cfg->in->TryPop())) {
		if (msg->type==MSG_STORE) {
			smsg=(Message_Store*)msg;
			aDebug(DEBUG_STORAGE, "ST->RADIUS/%c wr st:%u unit:%06X acct:%06X from:%u to:%u\n", smsg->prefix, s->instance, smsg->netunit, smsg->ap, smsg->data->from, smsg->ts);
			switch (smsg->prefix){
				case 'F':
					/*
					 * here RADIUS udp packet creation and forwarding code should go
					 * next line is an example of what we can send there
					 * fprintf(raw, "%u,%u,%u,%u,%llu,%llu\n", smsg->netunit, smsg->ap, smsg->data->from, smsg->ts, smsg->data->in, smsg->data->out);
					 */
					if( !radSendAccounting( smsg, cfg ) )
					    aLog(D_WARN, "Can not send RADIUS accounting request\n");
					else
					    num_raw++;
					break;
				case 'H':
				case 'D':
				case 'W':
				case 'M':
					/*
					 * skip other types
					 */
					break;
			}
		} else {
			aDebug(DEBUG_STORAGE, "message type %d not STORE - discarded\n",msg->type);
 		}
		msg=cfg->in->Pop();
	}

	if (num_raw) aDebug(DEBUG_STORAGE, "RADIUS->NET/raw %u messages\n", num_raw);
}

/*
    RADIUS code begins here. So, what attributes should we send?
    Standard attributes:
	User-Name
	NAS-IP-Address
	NAS-Port-Id
	NAS-Port-Type
	Acct-Status-Type
	Acct-Authentic
	Service-Type
	Acct-Session-Id
	Framed-IP-Address
	Acct-Input-Gigawords
	Acct-Output-Gigawords
	Acct-Input-Octets
	Acct-Output-Octets
	Acct-Session-Time
    VSA Attributes (NetAMS dictionary):
	Acct-Policy-Name
*/

#ifndef UINT32_MAX
#define	UINT32_MAX	0xFFFFFFFFU
#endif

int	radSendAccounting(Message_Store *msg, ServiceStorage_cfg *cfg) {
	struct rad_handle	*rh;
	int			is_error, attribute;
	char			session_id[33];
	char			oid[8];
	NetUnit			*unit;

	if((unit = Units->getUnitById(msg->netunit)) == NULL) {
	    aLog(D_ERR, "Unknown unit %06X\n", msg->netunit);
	    return 0;
	}
	
	if((rh = rad_acct_open()) == NULL) {
	    aLog(D_WARN, "Unable to construct accounting request\n");
	    return 0;
	}
	
	if(rad_add_server(rh, cfg->hostname?cfg->hostname:"localhost", cfg->port, cfg->password?cfg->password:"", cfg->timeout, cfg->retry) == -1) {
	    aLog(D_WARN, "Unable to configure servers\n");
	    rad_close( rh );
	    return 0;
	}

	if(rad_create_request(rh, RAD_ACCOUNTING_REQUEST) == -1) {
	    aLog(D_WARN, "Unable to create attributes space\n");
	    rad_close( rh );
	    return 0;
	}

	is_error = 1;
	while( 1 ) {
	    
	    /* basic info */
	    if(rad_put_int(rh, (attribute = RAD_ACCT_STATUS_TYPE), RAD_STOP)) break;
	    /* constructing session id */
	    sprintf(session_id, "%05u-%10lu-%s", cfg->pid, (u_long)msg->ts, unit->name);
    	    if(rad_put_string(rh, (attribute = RAD_ACCT_SESSION_ID), session_id)) break;

		// instead of name, put OID (name goes onto session id
	    sprintf(oid, "%06X", msg->netunit);
	    if(rad_put_string(rh, (attribute = RAD_USER_NAME), oid)) break;

		// for USERS and HOSTS, put IP address as optional parameter
		if (unit->type==NETUNIT_USER) {
			if(rad_put_addr(rh, (attribute = RAD_FRAMED_IP_ADDRESS), ((NetUnit_user*)unit)->ip)) break;
		}
		else if (unit->type==NETUNIT_HOST) {
			if(rad_put_addr(rh, (attribute = RAD_FRAMED_IP_ADDRESS), ((NetUnit_host*)unit)->ip)) break;
		}
		// type 55 is for Event-Timestamp (standard)
	    if(rad_put_int(rh, (attribute = RAD_EVENT_TIMESTAMP), msg->ts)) break;

	    // policy goes to Filter-Id
	    sprintf(oid, "%06X", msg->ap);
	    if(rad_put_string(rh, (attribute = RAD_FILTER_ID), oid)) break;

	    if(rad_put_addr(rh, (attribute = RAD_NAS_IP_ADDRESS), cfg->nas_ip)) break;
	    /* for the time present port is unknown for us */
	    if(rad_put_int(rh, (attribute = RAD_NAS_PORT), 0)) break;
	    
	    if(rad_put_int(rh, (attribute = RAD_SERVICE_TYPE), RAD_FRAMED)) break;
	    /* Do we really need Framed-Protocol???
	    if(rad_put_int(rh, (attribute = RAD_FRAMED_PROTOCOL), RAD_PPP)) break;
	    */
	    if(rad_put_int(rh, (attribute = RAD_NAS_PORT_TYPE), RAD_VIRTUAL)) break;
	    if(rad_put_int(rh, (attribute = RAD_ACCT_TERMINATE_CAUSE), RAD_TERM_HOST_REQUEST)) break;

	    /* session time is insignificant for us */
	    if(rad_put_int(rh, (attribute = RAD_ACCT_SESSION_TIME), (u_int32_t)(msg->ts - msg->data->from)) == -1) break;
	    if(rad_put_int(rh, (attribute = RAD_ACCT_INPUT_GIGAWORDS), (u_int32_t)(msg->data->in / UINT32_MAX)) == -1) break;
	    if(rad_put_int(rh, (attribute = RAD_ACCT_OUTPUT_GIGAWORDS), (u_int32_t)(msg->data->out / UINT32_MAX)) == -1) break;
	    if(rad_put_int(rh, (attribute = RAD_ACCT_INPUT_OCTETS), (u_int32_t)(msg->data->in % UINT32_MAX)) == -1) break;
	    if(rad_put_int(rh, (attribute = RAD_ACCT_OUTPUT_OCTETS), (u_int32_t)(msg->data->out % UINT32_MAX)) == -1) break;
	
	    if(rad_send_request(rh) != RAD_ACCOUNTING_RESPONSE)
		aLog(D_ERR, "Can not send radius request for OID %06X\n", msg->netunit);
	
	    is_error = 0;
	    break;
	}
	
	if( is_error )
	    aLog(D_ERR, "Can not store attribute %d\n", attribute);

	rad_close( rh );
	return 1;
}

#endif
