/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2005 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: users.c,v 1.28.2.3 2005/12/15 08:13:51 jura Exp $ */

#include "netams.h"

//////////////////////////////////////////////////////////////////////////
UsersList *Users;
User *uInternal;
//////////////////////////////////////////////////////////////////////////
User::User(char *n, u_char h){
	name=set_string(n);
	real_name=email=password=crypted=NULL;
	next=NULL;
	permissions=UPERM_NONE;
	hidden=h;
	id=0;
	mode=USER_HUMAN;
}

User::~User(){
	if(name) aFree(name);
	if(real_name) aFree(real_name);
	if(email) aFree(email);
	if(password) aFree(password);
	if(crypted) aFree(crypted);
}

void User::setName(char *s) { 
	if(name) aFree(name); 
	name=set_string(s);
	Users->changed_user=1; 
}
void User::setRealName(char *s) { 
	if(real_name) aFree(real_name); 
	real_name=set_string(s);
}
void User::setEmail(char *s) { 
	if(email) aFree(email); 
	email=set_string(s);
}
void User::setPassword(char *s) { 
	if(password) aFree(password); 
	// with new password encryption technology we must do following:
	// if there is plain text password given here, free old password string variable,
	// next prepare crypted form, and call setCrypted
	char *t=crypt(s, "$1$");
	if (t) setCrypted(t);
	Users->changed_pw=1;  
}

void User::setCrypted(char *s) { 
	if(crypted) aFree(crypted); 
	crypted=set_string(s);
}

void User::setPermissions(char *s) { 
	for (unsigned short i=0; i<256; i++) 
		if (permissions_list[i] && s && !strcmp(s, permissions_list[i])) permissions=i;
}

u_char User::getPermissions() { return permissions; }
//////////////////////////////////////////////////////////////////////////
UsersList::UsersList(){
	root=last=NULL;
	num_users=0;
	changed_pw=changed_user=1;
	netams_mutex_init(&lock, NULL);
}

UsersList::~UsersList(){
	User *u,*tmp=NULL;

	for(u=root;u!=NULL;u=tmp) {
		tmp=u->next;
		delete u;
	}
	
	netams_mutex_destroy(&lock);
}

void UsersList::Insert(User *s){
	netams_mutex_lock(&lock);
	User *d;
	for(d=root; d!=NULL; d=d->next)
		if (d==s) { 
			netams_mutex_unlock(&lock);
			return;
		}
	if (root==NULL) root=s;
	else last->next=s;
	last=s; 
	num_users++;
	netams_mutex_unlock(&lock);
}	

void UsersList::Delete(char *p){
	User *s = getUser(p);
	if (s) Delete(s);
}

void UsersList::Delete(User *s){
	netams_mutex_lock(&lock);
	User *d, *p=NULL;
	for(d=root; d!=NULL; d=d->next)	{
		if (d==s) {
			if (s==root && s==last ) root=last=NULL;
			else if (s==root) root=s->next;
			else if (s==last) { last=p; last->next=NULL; }
			else p->next=s->next;

			delete s;
			num_users--;
			break;
		}
		p=d; 
 	}
	netams_mutex_unlock(&lock);
}

User *UsersList::getUser(char *name) {
	netams_mutex_lock(&lock);
	User *d=NULL;
	for(d=root; d!=NULL; d=d->next)
		if (!strcmp(d->name, name)) 
			break;
	netams_mutex_unlock(&lock);
	
	if(!d) {
		oid id=strtol(name, NULL, 16);
		if(id) d=getUserById(id);
	}
	return d;
}

User *UsersList::getUserById(oid id){
	netams_mutex_lock(&lock);
	User *d=NULL;
	for(d=root; d!=NULL; d=d->next)
		if (d->id == id) 
			break;
	netams_mutex_unlock(&lock);
	return d;
}

void UsersList::listUsers(Connection *conn){
	netams_mutex_lock(&lock);
	User *t;
	fprintf(conn->stream_w, "%6s | %4s | %10s | %20s | %10s\n", "OID", "MODE", "NAME", "REAL NAME", "PERMIT");
	for(t=root; t!=NULL; t=t->next){
		if (!t->hidden) {
			fprintf(conn->stream_w, "%06X |", t->id);
			switch (t->mode) {
				case USER_HUMAN:		fprintf(conn->stream_w, " %4s |", "U"); break; 
				case USER_DATASOURCE:	fprintf(conn->stream_w, " %4s |", "DS"); break; 
				case USER_STORAGE:		fprintf(conn->stream_w, " %4s |", "ST"); break; 
				default: 				fprintf(conn->stream_w, " %4s |", "<\?\?>"); break; 
			}
			fprintf(conn->stream_w, " %10s | %20s | %10s\n", t->name, t->real_name?t->real_name:"<>", permissions_list[t->permissions]);
		}
	}
	netams_mutex_unlock(&lock);
}

void UsersList::listUsersCfg(FILE *f, u_char flags){
	netams_mutex_lock(&lock);
	User *t;
	for(t=root; t!=NULL; t=t->next){
		if (!t->hidden) {
			fprintf(f, "user oid %06X", t->id);
			if (t->name) fprintf(f, " name %s", t->name);
			if (t->real_name) fprintf(f, " real-name \"%s\"", t->real_name);
			if (t->password) fprintf(f, " password %s", (flags&CFG_NO_PASSWORDS)?"***":t->password);
			if (t->crypted) fprintf(f, " crypted %s", (flags&CFG_NO_PASSWORDS)?"***":t->crypted);
			if (t->email) fprintf(f, " email %s", t->email);
			fprintf(f, " permit %s\n", permissions_list[t->permissions]);
		}
	}
	netams_mutex_unlock(&lock);
}

void UsersList::listUsersHtml(Connection *conn){
	netams_mutex_lock(&lock);
	User *t;
	for (t=root; t!=NULL; t=t->next)
		if (!t->hidden && t->permissions >= UPERM_SHOWSTAT_HIS) 
			fprintf(conn->stream_w, " %s", t->name);
	netams_mutex_unlock(&lock);
}

void UsersList::listPasswordsHtml(Connection *conn){
	netams_mutex_lock(&lock);
	User *t;
	for (t=root; t!=NULL; t=t->next)
		if (!t->hidden && t->permissions >= UPERM_SHOWSTAT_HIS) {
			if (t->password) fprintf(conn->stream_w, "%s:%s\n", t->name, crypt(t->password, "$1$"));
			if (t->crypted) fprintf(conn->stream_w, "%s:%s\n", t->name, t->crypted);
		}
	netams_mutex_unlock(&lock);
}

//////////////////////////////////////////////////////////////////////////
int cUser(Connection *conn, char *param[], u_char no_flag){
	User *u=NULL;
	oid id=0;
	
	if (!strcasecmp(param[1], "?")) { cHelp(conn, no_flag, "user"); return PARSE_OK; }
	else if (!strcmp(param[1], "name")) u=Users->getUser(param[2]);
	else if (!strcmp(param[1], "oid")) {
		id=strtol(param[2], NULL, 16);	
		u=Users->getUserById(id);
	}
	else { aParse(conn, "user command unknown\n"); return PARSE_OK; }
	
	if (!u && !no_flag) { 
		u= new User(param[2]); 
		u->id=newOid(id);
		Users->Insert(u); 
		aParse(conn, "user %s %06X created\n", u->name, u->id); 
	}

	if (u && no_flag){
		aParse(conn, "user %s %06X deleted\n", u->name, u->id);
		Users->Delete(u);
		return PARSE_OK;
	}

	u_char i=3;
	while (param[i]!=empty) {
	 	if (strcasecmp(param[i], "real-name")==0) { 
			u->setRealName(param[i+1]);
			aParse(conn, "user %s real-name set: %s\n", u->name, u->real_name);

		} else if (strcasecmp(param[i], "name")==0) { 
			u->setName(param[i+1]);
			aParse(conn, "user %s name set: %s\n", u->name, u->name);

		} else if (strcasecmp(param[i], "email")==0) { 
			u->setEmail(param[i+1]);
			aParse(conn, "user %s email set: %s\n", u->name, u->email);

		} else if (strcasecmp(param[i], "password")==0) { 
			u->setPassword(param[i+1]);
			aParse(conn, "user %s password set: %s\n", u->name, u->password?u->password:"<>");

		} else if (strcasecmp(param[i], "crypted")==0) { 
			u->setCrypted(param[i+1]);
			aParse(conn, "user %s crypted password set: %s\n", u->name, u->crypted);

		} else if (strcasecmp(param[i], "permit")==0) { 
			u->setPermissions(param[i+1]);
			aParse(conn, "user %s permissions set: %s\n", u->name, permissions_list[u->getPermissions()]);
			// conn->permissions=u->getPermissions();
		}
		else if (strcasecmp(param[i], "oid")==0) ;//do nothing 
		else aParse(conn, "user %s command unknown\n", u->name); 

		i=i+2;
	}
	return PARSE_OK;
}
//////////////////////////////////////////////////////////////////////////
int cShowUsers(Connection *conn){
	Users->listUsers(conn);
	return PARSE_OK;
}
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
